# import SimpleITK as sitk
from warnings import warn

import cv2
import numpy as np
import scipy.ndimage.interpolation as sci
import skimage.transform as skt
from scipy.ndimage import gaussian_filter, map_coordinates

"""
vol: [H, W, D(, C)]

x, y, z <--> H, W, D

you should give (H, W, D) form shape.

skimage interpolation notations:

order = 0: Nearest-Neighbor
order = 1: Bi-Linear (default)
order = 2: Bi-Quadratic
order = 3: Bi-Cubic
order = 4: Bi-Quartic
order = 5: Bi-Quintic

Interpolation behaves strangely when input of type int.
** Be sure to change volume and mask data type to float !!! **
"""


def rotate2d(
    img,
    angle,
    axes=(0, 1),
    reshape=False,
    interpolation=1,
    border_mode="constant",
    value=0,
):
    return sci.rotate(
        img,
        angle,
        axes,
        reshape=reshape,
        order=interpolation,
        mode=border_mode,
        cval=value,
    )


def shift(img, shift, interpolation=1, border_mode="constant", value=0):
    return sci.shift(img, shift, order=interpolation, mode=border_mode, cval=value)


def crop(img, x1, y1, z1, x2, y2, z2):
    height, width, depth = img.shape[:3]
    if x2 <= x1 or y2 <= y1 or z2 <= z1:
        raise ValueError
    if x1 < 0 or y1 < 0 or z1 < 0:
        raise ValueError
    if x2 > height or y2 > width or z2 > depth:
        img = pad(img, (x2, y2, z2))
        warn("image size smaller than crop size, pad by default.", UserWarning)

    return img[x1:x2, y1:y2, z1:z2]


def get_center_crop_coords(height, width, depth, crop_height, crop_width, crop_depth):
    x1 = (height - crop_height) // 2
    x2 = x1 + crop_height
    y1 = (width - crop_width) // 2
    y2 = y1 + crop_width
    z1 = (depth - crop_depth) // 2
    z2 = z1 + crop_depth
    return x1, y1, z1, x2, y2, z2


def center_crop(img, crop_height, crop_width, crop_depth):
    height, width, depth = img.shape[:3]
    if height < crop_height or width < crop_width or depth < crop_depth:
        raise ValueError
    x1, y1, z1, x2, y2, z2 = get_center_crop_coords(
        height, width, depth, crop_height, crop_width, crop_depth
    )
    img = img[x1:x2, y1:y2, z1:z2]
    return img


def get_random_crop_coords(
    height, width, depth, crop_height, crop_width, crop_depth, h_start, w_start, d_start
):
    x1 = int((height - crop_height) * h_start)
    x2 = x1 + crop_height
    y1 = int((width - crop_width) * w_start)
    y2 = y1 + crop_width
    z1 = int((depth - crop_depth) * d_start)
    z2 = z1 + crop_depth
    return x1, y1, z1, x2, y2, z2


def random_crop(img, crop_height, crop_width, crop_depth, h_start, w_start, d_start):
    height, width, depth = img.shape[:3]
    if height < crop_height or width < crop_width or depth < crop_depth:
        img = pad(img, (crop_width, crop_height, crop_depth))
        warn("image size smaller than crop size, pad by default.", UserWarning)
    x1, y1, z1, x2, y2, z2 = get_random_crop_coords(
        height,
        width,
        depth,
        crop_height,
        crop_width,
        crop_depth,
        h_start,
        w_start,
        d_start,
    )
    img = img[x1:x2, y1:y2, z1:z2]
    return img


def normalize(img, range_norm=True):
    if range_norm:
        mn = img.min()
        mx = img.max()
        img = (img - mn) / (mx - mn)
    mean = img.mean()
    std = img.std()
    denominator = np.reciprocal(std)
    img = (img - mean) * denominator
    return img


def pad(image, new_shape, border_mode="constant", value=0):
    """
    image: [H, W, D, C] or [H, W, D]
    new_shape: [H, W, D]
    """
    axes_not_pad = len(image.shape) - len(new_shape)

    old_shape = np.array(image.shape[: len(new_shape)])
    new_shape = np.array(
        [max(new_shape[i], old_shape[i]) for i in range(len(new_shape))]
    )

    difference = new_shape - old_shape
    pad_below = difference // 2
    pad_above = difference - pad_below

    pad_list = [list(i) for i in zip(pad_below, pad_above)] + [[0, 0]] * axes_not_pad

    if border_mode == "reflect":
        res = np.pad(image, pad_list, border_mode)
    elif border_mode == "constant":
        res = np.pad(image, pad_list, border_mode, constant_values=value)
    else:
        raise ValueError

    return res


def gaussian_noise(img, var, mean):
    return img + np.random.normal(mean, var, img.shape)


def resize(img, new_shape, interpolation=1):
    """
    img: [H, W, D, C] or [H, W, D]
    new_shape: [H, W, D]
    """
    return skt.resize(
        img,
        new_shape,
        order=interpolation,
        mode="constant",
        cval=0,
        clip=True,
        anti_aliasing=False,
    )


def rescale(img, scale, interpolation=1):
    """
    img: [H, W, D, C] or [H, W, D]
    scale: scalar float
    """
    return skt.rescale(
        img,
        scale,
        order=interpolation,
        mode="constant",
        cval=0,
        clip=True,
        channel_axis=-1,
        anti_aliasing=False,
    )
    """
    shape = [int(scale * i) for i in img.shape[:3]]
    return resize(img, shape, interpolation)
    """


def gamma_transform(img, gamma, eps=1e-7):
    mn = img.min()
    rng = img.max() - mn
    img = (img - mn) / (rng + eps)
    return np.power(img, gamma)


def elastic_transform_pseudo2D(
    img,
    alpha,
    sigma,
    alpha_affine,
    interpolation=cv2.INTER_LINEAR,
    border_mode=cv2.BORDER_REFLECT_101,
    value=None,
    random_state=42,
    approximate=False,
):
    """Elastic deformation of images as described in [Simard2003]_ (with modifications).
    Based on https://gist.github.com/erniejunior/601cdf56d2b424757de5

    .. [Simard2003] Simard, Steinkraus and Platt, "Best Practices for
         Convolutional Neural Networks applied to Visual Document Analysis", in
         Proc. of the International Conference on Document Analysis and
         Recognition, 2003.
    """
    random_state = np.random.RandomState(random_state)

    height, width, depth = img.shape[:3]

    # Random affine
    center_square = np.float32((height, width)) // 2
    square_size = min((height, width)) // 3
    alpha = float(alpha)
    sigma = float(sigma)
    alpha_affine = float(alpha_affine)

    pts1 = np.float32(
        [
            center_square + square_size,
            [center_square[0] + square_size, center_square[1] - square_size],
            center_square - square_size,
        ]
    )
    salt = random_state.uniform(-alpha_affine, alpha_affine, size=pts1.shape).astype(
        np.float32
    )
    pts2 = pts1 + salt
    matrix = cv2.getAffineTransform(pts1, pts2)

    # pseoudo 2D
    res = np.zeros_like(img)
    for d in range(depth):
        tmp = img[:, :, d]  # [H, W, C]
        tmp = cv2.warpAffine(
            tmp,
            M=matrix,
            dsize=(width, height),
            flags=interpolation,
            borderMode=border_mode,
            borderValue=value,
        )
        res[:, :, d] = tmp
    img = res

    if approximate:
        # Approximate computation smooth displacement map with a large enough kernel.
        # On large images (512+) this is approximately 2X times faster
        dx = random_state.rand(height, width).astype(np.float32) * 2 - 1
        cv2.GaussianBlur(dx, (17, 17), sigma, dst=dx)
        dx *= alpha

        dy = random_state.rand(height, width).astype(np.float32) * 2 - 1
        cv2.GaussianBlur(dy, (17, 17), sigma, dst=dy)
        dy *= alpha
    else:
        dx = np.float32(
            gaussian_filter((random_state.rand(height, width) * 2 - 1), sigma) * alpha
        )
        dy = np.float32(
            gaussian_filter((random_state.rand(height, width) * 2 - 1), sigma) * alpha
        )

    x, y = np.meshgrid(np.arange(width), np.arange(height))

    map_x = np.float32(x + dx)
    map_y = np.float32(y + dy)

    # pseoudo 2D
    res = np.zeros_like(img)
    for d in range(depth):
        tmp = img[:, :, d]  # [H, W, C]
        tmp = cv2.remap(
            tmp,
            map1=map_x,
            map2=map_y,
            interpolation=interpolation,
            borderMode=border_mode,
            borderValue=value,
        )
        res[:, :, d] = tmp
    img = res

    return img


"""
Later are coordinates-based 3D rotation and elastic transforms.
reference: https://github.com/MIC-DKFZ/batchgenerators
"""


def elastic_transform(
    img,
    sigmas,
    alphas,
    interpolation=1,
    border_mode="constant",
    value=0,
    random_state=42,
):
    """
    img: [H, W, D(, C)]
    """
    coords = generate_coords(img.shape[:3])
    coords = elastic_deform_coords(coords, sigmas, alphas, random_state)
    coords = recenter_coords(coords)
    if len(img.shape) == 4:
        num_channels = img.shape[3]
        res = []
        for channel in range(num_channels):
            res.append(
                map_coordinates(
                    img[:, :, :, channel],
                    coords,
                    order=interpolation,
                    mode=border_mode,
                    cval=value,
                )
            )
        return np.stack(res, -1)
    else:
        return map_coordinates(
            img, coords, order=interpolation, mode=border_mode, cval=value
        )


def generate_coords(shape):
    """
    coords: [n_dim=3, H, W, D]
    """
    tmp = tuple([np.arange(i) for i in shape])
    coords = np.array(np.meshgrid(*tmp, indexing="ij")).astype(float)
    for d in range(len(shape)):
        coords[d] -= ((np.array(shape).astype(float) - 1) / 2)[d]
    return coords


def elastic_deform_coords(coords, sigmas, alphas, random_state):
    random_state = np.random.RandomState(random_state)
    n_dim = coords.shape[0]
    if not isinstance(alphas, (tuple, list)):
        alphas = [alphas] * n_dim
    if not isinstance(sigmas, (tuple, list)):
        sigmas = [sigmas] * n_dim
    offsets = []
    for d in range(n_dim):
        offset = gaussian_filter(
            (random_state.rand(*coords.shape[1:]) * 2 - 1),
            sigmas,
            mode="constant",
            cval=0,
        )
        mx = np.max(np.abs(offset))
        offset = alphas[d] * offset / mx
        offsets.append(offset)
    offsets = np.array(offsets)
    coords += offsets
    return coords


def recenter_coords(coords):
    n_dim = coords.shape[0]
    mean = coords.mean(axis=tuple(range(1, len(coords.shape))), keepdims=True)
    coords -= mean
    for d in range(n_dim):
        ctr = int(np.round(coords.shape[d + 1] / 2))
        coords[d] += ctr
    return coords


def rotate3d(img, x, y, z, interpolation=1, border_mode="constant", value=0):
    """
    img: [H, W, D(, C)]
    x, y, z: angle in degree.
    """
    x, y, z = [np.pi * i / 180 for i in [x, y, z]]
    coords = generate_coords(img.shape[:3])
    coords = rotate_coords(coords, x, y, z)
    coords = recenter_coords(coords)
    if len(img.shape) == 4:
        num_channels = img.shape[3]
        res = []
        for channel in range(num_channels):
            res.append(
                map_coordinates(
                    img[:, :, :, channel],
                    coords,
                    order=interpolation,
                    mode=border_mode,
                    cval=value,
                )
            )
        return np.stack(res, -1)
    else:
        return map_coordinates(
            img, coords, order=interpolation, mode=border_mode, cval=value
        )


def rotate_coords(coords, angle_x, angle_y, angle_z):
    rot_matrix = np.identity(len(coords))
    rot_matrix = rot_matrix @ rot_x(angle_x)
    rot_matrix = rot_matrix @ rot_y(angle_y)
    rot_matrix = rot_matrix @ rot_z(angle_z)
    coords = (
        np.dot(coords.reshape(len(coords), -1).transpose(), rot_matrix)
        .transpose()
        .reshape(coords.shape)
    )
    return coords


def rot_x(angle):
    rotation_x = np.array(
        [
            [1, 0, 0],
            [0, np.cos(angle), -np.sin(angle)],
            [0, np.sin(angle), np.cos(angle)],
        ]
    )
    return rotation_x


def rot_y(angle):
    rotation_y = np.array(
        [
            [np.cos(angle), 0, np.sin(angle)],
            [0, 1, 0],
            [-np.sin(angle), 0, np.cos(angle)],
        ]
    )
    return rotation_y


def rot_z(angle):
    rotation_z = np.array(
        [
            [np.cos(angle), -np.sin(angle), 0],
            [np.sin(angle), np.cos(angle), 0],
            [0, 0, 1],
        ]
    )
    return rotation_z


def rescale_warp(img, scale, interpolation=1, border_mode="constant", value=0):
    """
    img: [H, W, D(, C)]
    """
    coords = generate_coords(img.shape[:3])
    coords = scale_coords(coords, scale)
    coords = recenter_coords(coords)
    if len(img.shape) == 4:
        num_channels = img.shape[3]
        res = []
        for channel in range(num_channels):
            res.append(
                map_coordinates(
                    img[:, :, :, channel],
                    coords,
                    order=interpolation,
                    mode=border_mode,
                    cval=value,
                )
            )
        return np.stack(res, -1)
    else:
        return map_coordinates(
            img, coords, order=interpolation, mode=border_mode, cval=value
        )


def scale_coords(coords, scale):
    if isinstance(scale, (tuple, list, np.ndarray)):
        assert len(scale) == len(coords)
        for i in range(len(scale)):
            coords[i] *= scale[i]
    else:
        coords *= scale
    return coords



from .config import DataGenerationConfig
from .data_generator import DataGenerator

# +------------------------------------------------------------------------------------------+ #
# |                                  GENERATE SYNTHETIC DATA                                 | #
# +------------------------------------------------------------------------------------------+ #


def init_base_config():
    config = DataGenerationConfig()
    # outliers_pass1 = Outliers(
    #    radius_range_xy=(25, 80),
    #    radius_range_z=(20, 40),
    #    nb_points_range=(1000, 3000),
    #    nb_clusters_range=(100, 150),
    #    intensity_range=(0.5, 0.7),
    # )
    # outliers_pass2 = cfg.Outliers(
    #    radius_range_xy=(150, 300),
    #    radius_range_z=(20, 40),
    #    nb_points_range=(500, 1000),
    #    nb_clusters_range=(100, 150),
    #    intensity_range=(0.5, 1.),
    # )

    # Tilt
    # outliers_pass2 = Outliers(
    #    radius_range_xy=(150, 300),
    #    radius_range_z=(20, 40),
    #    nb_points_range=(500, 1000),
    #    nb_clusters_range=(100, 200),
    #    intensity_range=(0.5, 1.0),
    # )

    # config.outliers = (outliers_pass1, outliers_pass2) # no outlier
    return config


def generate_data(
    size: int, output_dir: str, pointcloud_path: str = None, extension: str = "npz"
) -> None:
    config = init_base_config()

    # Small image
    config.target_shape = (128, 512, 512)
    config.voxelisation.num_particles = 150
    config.voxelisation.max_particle_dim = 25

    # Tilt config
    config.voxelisation.tilt_margin = 0
    config.voxelisation.tilt_strategy = "uniform"
    config.voxelisation.bandwidth = 15
    config.voxelisation.cluster_range_xy = (10, 40)
    config.voxelisation.cluster_range_z = (5, 20)
    config.voxelisation.nb_particles_per_cluster_range = (1, 1)  # no cluster
    config.augmentation.intensity_mean_ratio = 0.0  # no hole
    config.augmentation.intensity_std_ratio_range = (0.0, 0.0)  # no hole
    config.augmentation.shrink_range = (1.0, 1.0)  # no shrink

    # sensors
    config.sensor.anisotropic_blur = False
    config.sensor.poisson_noise = False
    config.sensor.gaussian_noise = False

    config.io.generated_dataset_size = size
    config.io.output_dir = output_dir
    config.io.extension = extension

    DataGenerator(config).generate_dataset()



import argparse
import os

from .annotate import annotate


def parse_args() -> argparse.Namespace:
    """
    Arguments:
     - file: path to the input file
     - output: path to the output file
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("file", type=str, help="path to the image to annotate")
    parser.add_argument("output", type=str, help="path to the output csv file")
    parser.add_argument(
        "--spacing", type=float, default=None, nargs="+", help="Voxel size (ZYX)"
    )
    return parser.parse_args()


def main(args: argparse.Namespace) -> None:
    p, _ = os.path.splitext(args.output)
    if args.spacing is None:
        spacing = (1.0, 1.0, 1.0)
    else:
        assert len(args.spacing) == 3
        spacing = args.spacing
    annotate(args.file, p + ".csv", spacing=spacing)


if __name__ == "__main__":
    main(parse_args())



import random

import cv2
import numpy as np

from ..core.transforms_interface import DualTransform, Transform
from . import functionals as F


class Float(DualTransform):
    def apply(self, img):
        return img.astype(np.float32)


class Contiguous(DualTransform):
    def apply(self, img):
        return np.ascontiguousarray(img)


class PadIfNeeded(DualTransform):
    def __init__(
        self,
        shape,
        border_mode="constant",
        value=0,
        mask_value=0,
        always_apply=False,
        p=1,
    ):
        super().__init__(always_apply, p)
        self.shape = shape
        self.border_mode = border_mode
        self.value = value
        self.mask_value = mask_value

    def apply(self, img):
        return F.pad(img, self.shape, self.border_mode, self.value)

    def apply_to_mask(self, mask):
        return F.pad(mask, self.shape, self.border_mode, self.mask_value)


class GaussianNoise(Transform):
    def __init__(self, var_limit=(0, 0.1), mean=0, always_apply=False, p=0.5):
        super().__init__(always_apply, p)
        self.var_limit = var_limit
        self.mean = mean

    def apply(self, img, var):
        return F.gaussian_noise(img, var=var, mean=self.mean)

    def get_params(self, **data):
        return {"var": random.uniform(*self.var_limit)}


class Resize(DualTransform):
    def __init__(self, shape, interpolation=3, always_apply=False, p=1):
        super().__init__(always_apply, p)
        self.shape = shape
        self.interpolation = interpolation

    def apply(self, img):
        return F.resize(img, new_shape=self.shape, interpolation=self.interpolation)

    def apply_to_mask(self, mask):
        return F.resize(mask, new_shape=self.shape, interpolation=0)


class RandomScale(DualTransform):
    def __init__(
        self, scale_limit=[0.9, 1.1], interpolation=3, always_apply=False, p=0.5
    ):
        super().__init__(always_apply, p)
        self.scale_limit = scale_limit
        self.interpolation = interpolation

    def get_params(self, **data):
        return {"scale": random.uniform(self.scale_limit[0], self.scale_limit[1])}

    def apply(self, img, scale):
        return F.rescale(img, scale, interpolation=self.interpolation)

    def apply_to_mask(self, mask, scale):
        return F.rescale(mask, scale, interpolation=0)


class RandomScale2(DualTransform):
    """
    TODO: compare speeds with version 1.
    """

    def __init__(
        self, scale_limit=[0.9, 1.1], interpolation=3, always_apply=False, p=0.5
    ):
        super().__init__(always_apply, p)
        self.scale_limit = scale_limit
        self.interpolation = interpolation

    def get_params(self, **data):
        return {"scale": random.uniform(self.scale_limit[0], self.scale_limit[1])}

    def apply(self, img, scale):
        return F.rescale_warp(img, scale, interpolation=self.interpolation)

    def apply_to_mask(self, mask, scale):
        return F.rescale_warp(mask, scale, interpolation=0)


class RotatePseudo2D(DualTransform):
    def __init__(
        self,
        axes=(0, 1),
        limit=(-90, 90),
        interpolation=3,
        border_mode="constant",
        value=0,
        mask_value=0,
        always_apply=False,
        p=0.5,
    ):
        super().__init__(always_apply, p)
        self.axes = axes
        self.limit = limit
        self.interpolation = interpolation
        self.border_mode = border_mode
        self.value = value
        self.mask_value = mask_value

    def apply(self, img, angle):
        return F.rotate2d(
            img,
            angle,
            axes=self.axes,
            reshape=False,
            interpolation=self.interpolation,
            border_mode=self.border_mode,
            value=self.value,
        )

    def apply_to_mask(self, mask, angle):
        return F.rotate2d(
            mask,
            angle,
            axes=self.axes,
            reshape=False,
            interpolation=0,
            border_mode=self.border_mode,
            value=self.mask_value,
        )

    def get_params(self, **data):
        return {"angle": random.uniform(self.limit[0], self.limit[1])}


class RandomRotate90(DualTransform):
    def __init__(self, axes=(0, 1), always_apply=False, p=0.5):
        super().__init__(always_apply, p)
        self.axes = axes

    def apply(self, img, factor):
        return np.rot90(img, factor, axes=self.axes)

    def get_params(self, **data):
        return {"factor": random.randint(0, 3)}


class Flip(DualTransform):
    def __init__(self, axis=0, always_apply=False, p=0.5):
        super().__init__(always_apply, p)
        self.axis = axis

    def apply(self, img):
        return np.flip(img, self.axis)


class Normalize(Transform):
    def __init__(self, range_norm=False, always_apply=True, p=1.0):
        super().__init__(always_apply, p)
        self.range_norm = range_norm

    def apply(self, img):
        return F.normalize(img, range_norm=self.range_norm)


class Transpose(DualTransform):
    def __init__(self, axes=(1, 0, 2), always_apply=False, p=0.5):
        super().__init__(always_apply, p)
        self.axes = axes

    def apply(self, img):
        return np.transpose(img, self.axes)


class CenterCrop(DualTransform):
    def __init__(self, shape, always_apply=False, p=1.0):
        super().__init__(always_apply, p)
        self.shape = shape

    def apply(self, img):
        return F.center_crop(img, self.shape[0], self.shape[1], self.shape[2])


class RandomResizedCrop(DualTransform):
    def __init__(
        self, shape, scale_limit=(0.8, 1.2), interpolation=3, always_apply=False, p=1.0
    ):
        super().__init__(always_apply, p)
        self.shape = shape
        self.scale_limit = scale_limit
        self.interpolation = interpolation

    def apply(self, img, scale=1, scaled_shape=None, h_start=0, w_start=0, d_start=0):
        if scaled_shape is None:
            scaled_shape = self.shape
        img = F.random_crop(
            img,
            scaled_shape[0],
            scaled_shape[1],
            scaled_shape[2],
            h_start,
            w_start,
            d_start,
        )
        return F.resize(img, new_shape=self.shape, interpolation=self.interpolation)

    def apply_to_mask(
        self, img, scale=1, scaled_shape=None, h_start=0, w_start=0, d_start=0
    ):
        if scaled_shape is None:
            scaled_shape = self.shape
        img = F.random_crop(
            img,
            scaled_shape[0],
            scaled_shape[1],
            scaled_shape[2],
            h_start,
            w_start,
            d_start,
        )
        return F.resize(img, new_shape=self.shape, interpolation=0)

    def get_params(self, **data):
        scale = random.uniform(self.scale_limit[0], self.scale_limit[1])
        scaled_shape = [int(scale * i) for i in self.shape]
        return {
            "scale": scale,
            "scaled_shape": scaled_shape,
            "h_start": random.random(),
            "w_start": random.random(),
            "d_start": random.random(),
        }


class RandomCrop(DualTransform):
    def __init__(self, shape, always_apply=False, p=1.0):
        super().__init__(always_apply, p)
        self.shape = shape

    def apply(self, img, h_start=0, w_start=0, d_start=0):
        return F.random_crop(
            img, self.shape[0], self.shape[1], self.shape[2], h_start, w_start, d_start
        )

    def get_params(self, **data):
        return {
            "h_start": random.random(),
            "w_start": random.random(),
            "d_start": random.random(),
        }


class CropNonEmptyMaskIfExists(DualTransform):
    def __init__(self, shape, always_apply=False, p=1.0):
        super().__init__(always_apply, p)
        self.height = shape[0]
        self.width = shape[1]
        self.depth = shape[2]

    def apply(self, img, x_min=0, y_min=0, z_min=0, x_max=0, y_max=0, z_max=0):
        return F.crop(img, x_min, y_min, z_min, x_max, y_max, z_max)

    def get_params(self, **data):
        mask = data["mask"]  # [H, W, D]
        mask_height, mask_width, mask_depth = mask.shape

        if mask.sum() == 0:
            x_min = random.randint(0, mask_height - self.height)
            y_min = random.randint(0, mask_width - self.width)
            z_min = random.randint(0, mask_depth - self.depth)
        else:
            non_zero = np.argwhere(mask)
            x, y, z = random.choice(non_zero)
            x_min = x - random.randint(0, self.height - 1)
            y_min = y - random.randint(0, self.width - 1)
            z_min = z - random.randint(0, self.depth - 1)
            x_min = np.clip(x_min, 0, mask_height - self.height)
            y_min = np.clip(y_min, 0, mask_width - self.width)
            z_min = np.clip(z_min, 0, mask_depth - self.depth)

        x_max = x_min + self.height
        y_max = y_min + self.width
        z_max = z_min + self.depth

        return {
            "x_min": x_min,
            "x_max": x_max,
            "y_min": y_min,
            "y_max": y_max,
            "z_min": z_min,
            "z_max": z_max,
        }


class ResizedCropNonEmptyMaskIfExists(DualTransform):
    def __init__(
        self, shape, scale_limit=(0.8, 1.2), interpolation=3, always_apply=False, p=1.0
    ):
        super().__init__(always_apply, p)
        self.shape = shape
        self.scale_limit = scale_limit
        self.interpolation = interpolation

    def apply(self, img, x_min=0, y_min=0, z_min=0, x_max=0, y_max=0, z_max=0):
        img = F.crop(img, x_min, y_min, z_min, x_max, y_max, z_max)
        return F.resize(img, self.shape, interpolation=self.interpolation)

    def apply_to_mask(self, img, x_min=0, y_min=0, z_min=0, x_max=0, y_max=0, z_max=0):
        img = F.crop(img, x_min, y_min, z_min, x_max, y_max, z_max)
        return F.resize(img, self.shape, interpolation=0)

    def get_params(self, **data):
        mask = data["mask"]  # [H, W, D]
        mask_height, mask_width, mask_depth = mask.shape

        scale = random.uniform(self.scale_limit[0], self.scale_limit[1])
        height, width, depth = [int(scale * i) for i in self.shape]

        if mask.sum() == 0:
            x_min = random.randint(0, mask_height - self.height)
            y_min = random.randint(0, mask_width - self.width)
            z_min = random.randint(0, mask_depth - self.depth)
        else:
            non_zero = np.argwhere(mask)
            x, y, z = random.choice(non_zero)
            x_min = x - random.randint(0, self.height - 1)
            y_min = y - random.randint(0, self.width - 1)
            z_min = z - random.randint(0, self.depth - 1)
            x_min = np.clip(x_min, 0, mask_height - self.height)
            y_min = np.clip(y_min, 0, mask_width - self.width)
            z_min = np.clip(z_min, 0, mask_depth - self.depth)

        x_max = x_min + self.height
        y_max = y_min + self.width
        z_max = z_min + self.depth

        return {
            "x_min": x_min,
            "x_max": x_max,
            "y_min": y_min,
            "y_max": y_max,
            "z_min": z_min,
            "z_max": z_max,
        }


class RandomGamma(Transform):
    def __init__(self, gamma_limit=(0.7, 1.5), eps=1e-7, always_apply=False, p=0.5):
        super().__init__(always_apply, p)
        self.gamma_limit = gamma_limit
        self.eps = eps

    def apply(self, img, gamma=1):
        return F.gamma_transform(img, gamma=gamma, eps=self.eps)

    def get_params(self, **data):
        return {"gamma": random.uniform(self.gamma_limit[0], self.gamma_limit[1])}


class ElasticTransformPseudo2D(DualTransform):
    def __init__(
        self,
        alpha=1000,
        sigma=50,
        alpha_affine=1,
        approximate=False,
        always_apply=False,
        p=0.5,
    ):
        super().__init__(always_apply, p)
        self.alpha = alpha
        self.sigma = sigma
        self.alpha_affine = alpha_affine
        self.approximate = approximate

    def apply(self, img, random_state=None):
        return F.elastic_transform_pseudo2D(
            img,
            self.alpha,
            self.sigma,
            self.alpha_affine,
            interpolation=cv2.INTER_LINEAR,
            border_mode=cv2.BORDER_REFLECT_101,
            value=None,
            random_state=random_state,
            approximate=False,
        )

    def apply_to_mask(self, img, random_state=None):
        return F.elastic_transform_pseudo2D(
            img,
            self.alpha,
            self.sigma,
            self.alpha_affine,
            interpolation=cv2.INTER_NEAREST,
            border_mode=cv2.BORDER_REFLECT_101,
            value=None,
            random_state=random_state,
            approximate=False,
        )

    def get_params(self, **data):
        return {"random_state": random.randint(0, 10000)}


class ElasticTransform(DualTransform):
    def __init__(
        self,
        deformation_limits=(0, 0.25),
        interpolation=3,
        border_mode="constant",
        value=0,
        mask_value=0,
        always_apply=False,
        p=0.5,
    ):
        super().__init__(always_apply, p)
        self.deformation_limits = deformation_limits
        self.interpolation = interpolation
        self.border_mode = border_mode
        self.value = value
        self.mask_value = mask_value

    def apply(self, img, sigmas, alphas, random_state=None):
        return F.elastic_transform(
            img,
            sigmas,
            alphas,
            interpolation=self.interpolation,
            random_state=random_state,
            border_mode=self.border_mode,
            value=self.value,
        )

    def apply_to_mask(self, img, sigmas, alphas, random_state=None):
        return F.elastic_transform(
            img,
            sigmas,
            alphas,
            interpolation=0,
            random_state=random_state,
            border_mode=self.border_mode,
            value=self.mask_value,
        )

    def get_params(self, **data):
        image = data["image"]  # [H, W, D]
        random_state = random.randint(0, 10000)
        deformation = random.uniform(*self.deformation_limits)
        sigmas = [deformation * x for x in image.shape[:3]]
        alphas = [random.uniform(x / 8, x / 2) for x in sigmas]
        return {
            "random_state": random_state,
            "sigmas": sigmas,
            "alphas": alphas,
        }


class Rotate(DualTransform):
    def __init__(
        self,
        x_limit=(-15, 15),
        y_limit=(-15, 15),
        z_limit=(-15, 15),
        interpolation=3,
        border_mode="constant",
        value=0,
        mask_value=0,
        always_apply=False,
        p=0.5,
    ):
        super().__init__(always_apply, p)
        self.x_limit = x_limit
        self.y_limit = y_limit
        self.z_limit = z_limit
        self.interpolation = interpolation
        self.border_mode = border_mode
        self.value = value
        self.mask_value = mask_value

    def apply(self, img, x, y, z):
        return F.rotate3d(
            img,
            x,
            y,
            z,
            interpolation=self.interpolation,
            border_mode=self.border_mode,
            value=self.value,
        )

    def apply_to_mask(self, mask, x, y, z):
        return F.rotate3d(
            mask,
            x,
            y,
            z,
            interpolation=0,
            border_mode=self.border_mode,
            value=self.mask_value,
        )

    def get_params(self, **data):
        return {
            "x": random.uniform(self.x_limit[0], self.x_limit[1]),
            "y": random.uniform(self.y_limit[0], self.y_limit[1]),
            "z": random.uniform(self.z_limit[0], self.z_limit[1]),
        }


class RemoveEmptyBorder(DualTransform):
    def __init__(self, border_value=0, always_apply=False, p=1.0):
        super().__init__(always_apply, p)
        self.border_value = border_value

    def apply(self, img, x_min=0, y_min=0, z_min=0, x_max=0, y_max=0, z_max=0):
        return F.crop(img, x_min, y_min, z_min, x_max, y_max, z_max)

    def get_params(self, **data):
        image = data["image"]  # [H, W, D, C]

        borders = np.where(image != self.border_value)

        return {
            "x_min": np.min(borders[0]),
            "x_max": np.max(borders[0]) + 1,
            "y_min": np.min(borders[1]),
            "y_max": np.max(borders[1]) + 1,
            "z_min": np.min(borders[2]),
            "z_max": np.max(borders[2]) + 1,
        }



# Some functions were taken from aicsimageio source code
from __future__ import annotations

import csv
import math
import os
import pickle
from collections import Counter
from typing import Any, Dict, Optional, Tuple

import imageio
import numpy as np
import tifffile
from ome_types import OME, from_xml
from ome_types.model.simple_types import UnitsLength

import spfluo
from spfluo.utils.volume import interpolate_to_size
from spfluo.utils.volume import resample as _resample


class DimensionNames:
    Time = "T"
    Channel = "C"
    SpatialZ = "Z"
    SpatialY = "Y"
    SpatialX = "X"
    Samples = "S"
    MosaicTile = "M"


DEFAULT_DIMENSION_ORDER_LIST = [
    DimensionNames.Time,
    DimensionNames.Channel,
    DimensionNames.SpatialZ,
    DimensionNames.SpatialY,
    DimensionNames.SpatialX,
]

DEFAULT_DIMENSION_ORDER = "".join(DEFAULT_DIMENSION_ORDER_LIST)

DEFAULT_DIMENSION_ORDER_LIST_WITH_SAMPLES = DEFAULT_DIMENSION_ORDER_LIST + [
    DimensionNames.Samples
]

DEFAULT_DIMENSION_ORDER_WITH_SAMPLES = "".join(
    DEFAULT_DIMENSION_ORDER_LIST_WITH_SAMPLES
)


def transpose_to_dims(
    data: np.ndarray,
    given_dims: str,
    return_dims: str,
) -> np.ndarray:
    """
    This shuffles the data dimensions from given_dims to return_dims. Each dimension
    must be present in given_dims must be used in return_dims

    Parameters
    ----------
    data: types.ArrayLike
        Either a dask array or numpy.ndarray of arbitrary shape but with the dimensions
        specified in given_dims
    given_dims: str
        The dimension ordering of data, "CZYX", "VBTCXZY" etc
    return_dims: str
        The dimension ordering of the return data

    Returns
    -------
    data: types.ArrayLike
        The data with the specified dimension ordering.

    Raises
    ------
    ConflictingArgumentsError
        given_dims and return_dims are incompatible.
    """
    # Use a counter to track that the contents are composed of the same letters
    # and that no letter is repeated
    if (
        Counter(given_dims) != Counter(return_dims)
        or max(Counter(given_dims).values()) > 1
    ):
        raise ValueError(
            f"given_dims={given_dims} and return_dims={return_dims} are incompatible."
        )

    # Resort the data into return_dims order
    match_map = {dim: given_dims.find(dim) for dim in given_dims}
    transposer = []
    for dim in return_dims:
        transposer.append(match_map[dim])
    data = data.transpose(transposer)

    return data


def reduce_to_slice(L: list | tuple) -> int | list | slice | tuple:
    # if the list only has one element, then just use it
    if len(L) == 1:
        return slice(L[0], L[0] + 1)
    # if the list has at least 2 elements we can check for sliceable
    # it is convertable to a slice if the step size between each
    # consecutive pair of elements is equal and positive
    # 1. get all the deltas in a list:
    steps = [(L[i + 1] - L[i]) for i in range(len(L) - 1)]
    # 2. check if all the deltas are equal and positive
    if steps[0] > 0 and steps.count(steps[0]) == len(steps):
        return slice(min(L), max(L) + 1, steps[0])
    # if we can't convert to a slice, then just return the list unmodified
    return L


def reshape_data(
    data: np.ndarray, given_dims: str, return_dims: str, **kwargs: Any
) -> np.ndarray:
    """
    Reshape the data into return_dims, pad missing dimensions, and prune extra
    dimensions. Warns the user to use the base reader if the depth of the Dimension
    being removed is not 1.

    Parameters
    ----------
    data: types.ArrayLike
        Either a dask array or numpy.ndarray of arbitrary shape but with the dimensions
        specified in given_dims
    given_dims: str
        The dimension ordering of data, "CZYX", "VBTCXZY" etc
    return_dims: str
        The dimension ordering of the return data
    kwargs:
        * C=1 => desired specific channel, if C in the input data has depth 3 then C=1
          returns the 2nd slice (0 indexed)
        * Z=10 => desired specific channel, if Z in the input data has depth 20 then
          Z=10 returns the 11th slice
        * T=[0, 1] => desired specific timepoints, if T in the input data has depth 100
          then T=[0, 1] returns the 1st and 2nd slice (0 indexed)
        * T=(0, 1) => desired specific timepoints, if T in the input data has depth 100
          then T=(0, 1) returns the 1st and 2nd slice (0 indexed)
        * T=(0, -1) => desired specific timepoints, if T in the input data has depth 100
          then T=(0, -1) returns the first and last slice
        * T=range(10) => desired specific timepoints, if T in the input data has depth
          100 then T=range(10) returns the first ten slices
        * T=slice(0, -1, 5) => desired specific timepoints, T=slice(0, -1, 5) returns
          every fifth timepoint

    Returns
    -------
    data: types.ArrayLike
        The data with the specified dimension ordering.

    Raises
    ------
    ConflictingArgumentsError
        Missing dimension in return dims when using range, slice, or multi-index
        dimension selection for the requested dimension.

    IndexError
        Requested dimension index not present in data.

    Examples
    --------
    Specific index selection

    >>> data = np.random.rand((10, 100, 100))
    ... z1 = reshape_data(data, "ZYX", "YX", Z=1)

    List of index selection

    >>> data = np.random.rand((10, 100, 100))
    ... first_and_second = reshape_data(data, "ZYX", "YX", Z=[0, 1])

    Tuple of index selection

    >>> data = np.random.rand((10, 100, 100))
    ... first_and_last = reshape_data(data, "ZYX", "YX", Z=(0, -1))

    Range of index selection

    >>> data = np.random.rand((10, 100, 100))
    ... first_three = reshape_data(data, "ZYX", "YX", Z=range(3))

    Slice selection

    >>> data = np.random.rand((10, 100, 100))
    ... every_other = reshape_data(data, "ZYX", "YX", Z=slice(0, -1, 2))

    Empty dimension expansion

    >>> data = np.random.rand((10, 100, 100))
    ... with_time = reshape_data(data, "ZYX", "TZYX")

    Dimension order shuffle

    >>> data = np.random.rand((10, 100, 100))
    ... as_zx_base = reshape_data(data, "ZYX", "YZX")

    Selections, empty dimension expansions, and dimension order shuffle

    >>> data = np.random.rand((10, 100, 100))
    ... example = reshape_data(data, "CYX", "BSTCZYX", C=slice(0, -1, 3))
    """
    # Check for parameter conflicts
    for dim in given_dims:
        # return_dims='CZYX' and iterable dimensions 'T=range(10)'
        # Dimension is in kwargs
        # Dimension is an iterable
        # Dimension is not in return dimensions
        if (
            isinstance(kwargs.get(dim), (list, tuple, range, slice))
            and dim not in return_dims
        ):
            raise ValueError(
                f"When selecting a multiple dimension indices, the specified "
                f"dimension must be provided in return_dims. "
                f"return_dims={return_dims}, dimension {dim} = {kwargs.get(dim)}"
            )

    # Process each dimension available
    new_dims = given_dims
    dim_specs = []
    for dim in given_dims:
        # Store index of the dim as it is in given data
        dim_index = given_dims.index(dim)

        # Handle dim in return_dims which means that it is
        # an iterable or None selection
        if dim in return_dims:
            # Specific iterable requested
            if dim in kwargs:
                # Actual dim specification
                # The specification provided for this dimension in the kwargs
                dim_spec = kwargs.get(dim)
                display_dim_spec = dim_spec

                if isinstance(dim_spec, int):
                    dim_spec = slice(dim_spec, dim_spec + 1)

                # Convert operator to standard list or slice
                # dask.Array and numpy.ndarray both natively support
                # List[int] and slices being passed to getitem so no need to cast them
                # to anything different
                if isinstance(dim_spec, (tuple, range)):
                    dim_spec = list(dim_spec)

                # Get the largest absolute value index in the list using min and max
                if isinstance(dim_spec, list):
                    check_selection_max = max([abs(min(dim_spec)), max(dim_spec)])
                    # try to convert to slice if possible
                    dim_spec = reduce_to_slice(dim_spec)

                # Get the largest absolute value index from start and stop of slice
                if isinstance(dim_spec, slice):
                    check_selection_max = max([abs(dim_spec.stop), abs(dim_spec.start)])
            else:
                # Nothing was requested from this dimension
                dim_spec = slice(None, None, None)
                display_dim_spec = dim_spec

                # No op means that it doesn't matter how much data is in this dimension
                check_selection_max = 0

        # Not in return_dims means that it is a fixed integer selection
        else:
            if dim in kwargs:
                # Integer requested
                dim_spec = kwargs.get(dim)
                display_dim_spec = dim_spec

                # Check that integer
                check_selection_max = dim_spec
            else:
                dim_spec = 0
                display_dim_spec = dim_spec
                check_selection_max = 0

            # Remove dim from new dims as it is fixed size
            new_dims = new_dims.replace(dim, "")

        # Check that fixed integer request isn't outside of request
        if check_selection_max > data.shape[dim_index]:
            raise IndexError(
                f"Dimension specified with {dim}={display_dim_spec} "
                f"but Dimension shape is {data.shape[dim_index]}."
            )

        # All checks and operations passed, append dim operation to getitem ops
        dim_specs.append(dim_spec)

    # Run getitems
    data = data[tuple(dim_specs)]

    # Add empty dims where dimensions were requested but data doesn't exist
    # Add dimensions to new dims where empty dims are added
    for i, dim in enumerate(return_dims):
        # This dimension wasn't processed
        if dim not in given_dims:
            new_dims = f"{new_dims[:i]}{dim}{new_dims[i:]}"
            data = data.reshape(*data.shape[:i], 1, *data.shape[i:])

    # Any extra dimensions have been removed, only a problem if the depth is > 1
    return transpose_to_dims(
        data, given_dims=new_dims, return_dims=return_dims
    )  # don't pass kwargs or 2 copies


def _get_dims_from_ome(ome: OME, scene_index: int) -> list[str]:
    """
    Process the OME metadata to retrieve the dimension names.

    Parameters
    ----------
    ome: OME
        A constructed OME object to retrieve data from.
    scene_index: int
        The current operating scene index to pull metadata from.

    Returns
    -------
    dims: List[str]
        The dimension names pulled from the OME metadata.

    Taken from aicsimageio
    """
    # Select scene
    scene_meta = ome.images[scene_index]

    # Create dimension order by getting the current scene's dimension order
    # and reversing it because OME store order vs use order is :shrug:
    dims = [d for d in scene_meta.pixels.dimension_order.value[::-1]]

    # Check for num samples and expand dims if greater than 1
    n_samples = scene_meta.pixels.channels[0].samples_per_pixel
    if n_samples is not None and n_samples > 1 and "S" not in dims:
        # Append to the end, i.e. the last dimension
        dims.append("S")

    return dims


def _guess_ome_dim_order(
    tiff: tifffile.TiffFile, ome: OME, scene_index: int
) -> list[str]:
    """
    Guess the dimension order based on OME metadata and actual TIFF data.
    Parameters
    -------
    tiff: TiffFile
        A constructed TIFF object to retrieve data from.
    ome: OME
        A constructed OME object to retrieve data from.
    scene_index: int
        The current operating scene index to pull metadata from.
    Returns
    -------
    dims: List[str]
        Educated guess of the dimension order for the file

    Taken from aicsimageio
    """
    dims_from_ome = _get_dims_from_ome(ome, scene_index)

    # Assumes the dimensions coming from here are align semantically
    # with the dimensions specified in this package. Possible T dimension
    # is not equivalent to T dimension here. However, any dimensions
    # not also found in OME will be omitted.
    dims_from_tiff_axes = list(tiff.series[scene_index].axes)

    # Adjust the guess of what the dimensions are based on the combined
    # information from the tiff axes and the OME metadata.
    # Necessary since while OME metadata should be source of truth, it
    # does not provide enough data to guess which dimension is Samples
    # for RGB files
    dims = [dim for dim in dims_from_ome if dim not in dims_from_tiff_axes]
    dims += [dim for dim in dims_from_tiff_axes if dim in dims_from_ome]
    return dims


def _expand_dims_to_match_ome(
    image_data: np.ndarray,
    ome: OME,
    dims: list[str],
    scene_index: int,
) -> np.ndarray:
    # Expand image_data for empty dimensions
    ome_shape = []

    # need to correct channel count if this is a RGB image
    n_samples = ome.images[scene_index].pixels.channels[0].samples_per_pixel
    has_multiple_samples = n_samples is not None and n_samples > 1
    for d in dims:
        # SizeC can represent RGB (Samples) data rather
        # than channel data, whether or not this is the case depends
        # on what the SamplesPerPixel are for the channel
        if d == "C" and has_multiple_samples:
            count = len(ome.images[scene_index].pixels.channels)
        elif d == "S" and has_multiple_samples:
            count = n_samples
        else:
            count = getattr(ome.images[scene_index].pixels, f"size_{d.lower()}")
        ome_shape.append(count)

    # The file may not have all the data but OME requires certain dimensions
    # expand to fill
    expand_dim_ops: list[Optional[slice]] = []
    for d_size in ome_shape:
        # Add empty dimension where OME requires dimension but no data exists
        if d_size == 1:
            expand_dim_ops.append(None)
        # Add full slice where data exists
        else:
            expand_dim_ops.append(slice(None, None, None))

    # Apply operators to dask array
    return image_data[tuple(expand_dim_ops)]


def _general_data_array_constructor(
    image_data: np.ndarray,
    ome: OME,
    scene_index: int,
    dims: list[str],
):
    # Expand the image data to match the OME empty dimensions
    image_data = _expand_dims_to_match_ome(
        image_data=image_data,
        ome=ome,
        dims=dims,
        scene_index=scene_index,
    )

    # Always order array
    if DimensionNames.Samples in dims:
        out_order = DEFAULT_DIMENSION_ORDER_WITH_SAMPLES
    else:
        out_order = DEFAULT_DIMENSION_ORDER

    # Transform into order
    image_data = reshape_data(
        image_data,
        "".join(dims),
        out_order,
    )

    # Reset dims after transform
    dims = [d for d in out_order]

    return image_data, dims


def get_data_from_ome_tiff(
    tiff: tifffile.TiffFile, scene_index: int, order: str = "CZYX"
):
    """returns data in the order asked"""
    assert tiff.is_ome
    dims = _guess_ome_dim_order(tiff, from_xml(tiff.ome_metadata), scene_index)
    image_data, dims = _general_data_array_constructor(
        tiff.series[scene_index].asarray(),
        from_xml(tiff.ome_metadata),
        scene_index,
        dims,
    )

    # Special rule:
    # if T>1 and C=1, transpose
    if (
        "C" in order
        and "T" not in order
        and image_data.shape[dims.index("T")] > 1
        and image_data.shape[dims.index("C")] == 1
    ):
        x = list(range(image_data.ndim))
        x[dims.index("T")] = dims.index("C")
        x[dims.index("C")] = dims.index("T")
        image_data = image_data.transpose(*tuple(x))

    unspecified_dims = set(dims).difference(set(order))
    for dim in unspecified_dims:
        assert (
            image_data.shape[dims.index(dim)] == 1
        ), f"the dimension {dim} is not 1 but {image_data.shape[dims.index(dim)]} "
        " in image of shape {image_data.shape}"
    image_data = image_data.squeeze(
        axis=tuple([dims.index(dim) for dim in unspecified_dims])
    )
    dims = [dim for dim in dims if dim not in unspecified_dims]
    assert len(order) == len(dims), f"{len(order)=} != {len(dims)=}"
    return image_data.transpose(*tuple([dims.index(o) for o in order]))


def get_cupy_array(image):
    if spfluo.has_cupy():
        import cupy as cp

        return cp.array(image)
    else:
        return image


def get_ndimage():
    if spfluo.has_cupy():
        from cupyx.scipy import ndimage
    else:
        from scipy import ndimage
    return ndimage


def get_numpy_array(image):
    if not isinstance(image, np.ndarray):
        return image.get()
    else:
        return image


def load_array(path: str) -> np.ndarray:
    """Takes a complete path to a file and return the corresponding numpy array.

    Args:
        path (str): Path to the file to read. It MUST contains the extension as this
                    will determines the way the numpy array is loaded from the file.

    Returns:
        np.ndarray: The array stored in the file described by 'path'.
    """
    extension = os.path.splitext(path)[-1]
    if extension == ".npz":
        return np.load(path)["image"]
    elif extension in [".tif", ".tiff"]:
        image = imageio.volread(path).astype(np.int16)
        # Some tiff images are heavily imbalanced:
        # their data type is int16 but very few voxels are actually > 255.
        # If this is the case, the image in truncated and casted to uint8.
        if image.dtype == np.int16 and ((image > 255).sum() / image.size) < 1e-3:
            image[image > 255] = 255
            image = image.astype(np.uint8)
        return image
    error_msg = f"Found extension {extension}. Extension must be one of npz or tif."
    raise NotImplementedError(error_msg)


def load_annotations(csv_path: str) -> np.ndarray:
    """Csv containing coordinates of objects center.

    Args:
        csv_path (str): Path of the file to read.

    Returns:
        np.ndarray: Array into which each line is alike
            ('image name', particle_id, z, y, x).
    """
    with open(csv_path, "r") as f:
        lines = f.readlines()
    data = []
    for line in lines:
        line_data = line.split(",")
        line_data[2:] = list(map(float, line_data[2:]))
        data.append(line_data)
    return np.array(data, dtype=object)


def load_angles(csv_path: str) -> np.ndarray:
    """Csv containing angles of particles.

    Args:
        csv_path (str): Path of the file to read.

    Returns:
        np.ndarray: Array into which each line is alike
            ('image name', particle_id, z, y, x).
    """
    with open(csv_path, "r") as f:
        lines = f.readlines()
    data = []
    for line in lines:
        line_data = line.split(",")
        line_data[2:] = list(map(float, line_data[2:]))
        data.append(line_data)
    return np.array(data, dtype=object)


def load_views(views_path, extension=None):
    _, ext = os.path.splitext(views_path)
    if ext == ".csv":  # DONT USE CSV NOT WORKING
        views_ = load_annotations(views_path)
        views = views_[:, 2].astype(int)
        if extension is None:
            patches_names = np.array(
                [
                    os.path.splitext(im_name)[0]
                    + "_"
                    + patch_index
                    + os.path.splitext(im_name)[1]
                    for im_name, patch_index in views_[:, [0, 1]]
                ]
            )
        else:
            patches_names = np.array(
                [
                    os.path.splitext(im_name)[0] + "_" + patch_index + extension
                    for im_name, patch_index in views_[:, [0, 1]]
                ]
            )

    elif ext == ".pickle":
        with open(views_path, "rb") as f:
            views_ = pickle.load(f)
        views = np.concatenate([views_[image_name][0] for image_name in views_.keys()])
        patches_names = np.concatenate(
            [views_[image_name][2] for image_name in views_.keys()]
        )

    return views, patches_names


def load_patches(crop_dir, patches_names):
    patches = [
        load_array(os.path.join(crop_dir, p)) for p in patches_names
    ]  # (N,z,y,x)
    return np.stack(patches, axis=0)


def load_pointcloud(pointcloud_path: str) -> np.ndarray:
    template_point_cloud = np.loadtxt(pointcloud_path, delimiter=",")
    return template_point_cloud


def center_to_corners(center: Tuple[int], size: Tuple[int]) -> Tuple[int]:
    depth, height, width = size
    center_z, center_y, center_x = center
    z_min = center_z - depth // 2
    y_min = center_y - height // 2
    x_min = center_x - width // 2
    z_max = z_min + depth
    y_max = y_min + height
    x_max = x_min + width
    return x_min, y_min, z_min, x_max, y_max, z_max


def summary(
    kwargs: Dict, title: str, output: str = None, return_table: bool = False
) -> None:
    # 1. Get key max length
    key_length = max([len(k) for k in kwargs.keys()])
    # 2. Get total length
    length = max([key_length + len(str(v)) for k, v in kwargs.items()]) + 4
    # 3. Define table delimiter and title
    hbar = "+" + length * "-" + "+" + "\n"
    pad_left = (length - len(title)) // 2
    pad_right = length - len(title) - pad_left
    title = "|" + pad_left * " " + title + pad_right * " " + "|" + "\n"
    # 4. Create table string
    table = "\n" + hbar + title + hbar
    # 5. Add lines to table
    for k, v in kwargs.items():
        line = k + (key_length - len(k)) * "." + ": " + str(v)
        line = "| " + line + (length - 1 - len(line)) * " " + "|" + "\n"
        table += line
    table += hbar
    # 6. Print table
    print(table)
    # 7. Save table to txt file (optional)
    if output is not None:
        with open(output, "w") as file:
            file.write(table)
    # 8. Return table (optional)
    if return_table:
        return table


def send_mail(subject: str, body: str) -> None:
    import smtplib
    import ssl

    port = 465  # SSL
    password = "hackitifyouwantidontcare"
    sender_email = "icuberemotedev@gmail.com"
    receiver_email = "vedrenneluc@gmail.com"
    message = f"""\
    {subject}

    {body}
    """
    context = ssl.create_default_context()
    with smtplib.SMTP_SSL("smtp.gmail.com", port, context=context) as server:
        server.login(sender_email, password)
        server.sendmail(sender_email, receiver_email, message)


def crop_one_particle(
    image: np.ndarray,
    center: np.ndarray,
    crop_size: Tuple[int],
    max_size: Tuple[int],
) -> np.ndarray:
    corners = center_to_corners(center, crop_size)
    corners = reframe_corners_if_needed(corners, crop_size, max_size)
    x_min, y_min, z_min, x_max, y_max, z_max = corners
    return image[z_min:z_max, y_min:y_max, x_min:x_max]


def reframe_corners_if_needed(
    corners: Tuple[int], crop_size: Tuple[int], max_size: Tuple[int]
) -> Tuple[int]:
    d, h, w = crop_size
    D, H, W = max_size
    x_min, y_min, z_min, x_max, y_max, z_max = corners
    z_min, y_min, x_min = max(0, z_min), max(0, y_min), max(0, x_min)
    # z_max, y_max, x_max = z_min + d, y_min + h, x_min + w
    z_max, y_max, x_max = min(D, z_max), min(H, y_max), min(W, x_max)
    # z_min, y_min, x_min = z_max - d, y_max - h, x_max - w
    return x_min, y_min, z_min, x_max, y_max, z_max


def resize(im_paths: str, size: float, folder_path: str):
    """pad isotropic OME-TIFF images to match the shape of a cube of size (size, size,
    size).
    If size isn't a multiple of the pixel size, new_size = ceil(size/pixel_size)
    Arguments:
        im_paths
        size
            shape of the cube in µm
        folder_path
            output folder
    """
    target_physical_size = size
    os.makedirs(folder_path, exist_ok=True)
    for im_path in im_paths:
        tif = tifffile.TiffFile(im_path, is_ome=True)
        ome = from_xml(tif.ome_metadata)
        im = get_data_from_ome_tiff(tif, 0, order="CZYX")

        assert len(ome.images) == 1
        # assert image is isotropic
        (pixel_physical_size,) = set(
            [
                ome.images[0].pixels.physical_size_x,
                ome.images[0].pixels.physical_size_y,
                ome.images[0].pixels.physical_size_z,
            ]
        )
        (pixel_physical_size_unit,) = set(
            [
                ome.images[0].pixels.physical_size_x_unit,
                ome.images[0].pixels.physical_size_y_unit,
                ome.images[0].pixels.physical_size_z_unit,
            ]
        )

        assert pixel_physical_size_unit == UnitsLength.MICROMETER
        pixel_size = math.ceil(target_physical_size / pixel_physical_size)
        im_resized = interpolate_to_size(im, (pixel_size,) * 3, multichannel=True)
        filename = os.path.join(folder_path, os.path.basename(im_path))

        # copy ome metadata to filename
        tifffile.imwrite(
            filename,
            im_resized,
            metadata={
                "axes": "CZYX",
                "PhysicalSizeX": pixel_physical_size,
                "PhysicalSizeXUnit": "µm",
                "PhysicalSizeY": pixel_physical_size,
                "PhysicalSizeYUnit": "µm",
                "PhysicalSizeZ": pixel_physical_size,
                "PhysicalSizeZUnit": "µm",
            },
        )
        tif.close()


def resample(im_paths: str, folder_path: str, target_pixel_physical_size: float = 1.0):
    """resample images to match the target physical size
    Args
        target_physical_size: float
            in µm
    """
    os.makedirs(folder_path, exist_ok=True)
    for im_path in im_paths:
        tif = tifffile.TiffFile(im_path, is_ome=True)
        image = get_data_from_ome_tiff(tif, 0, order="CZYX")
        ome = from_xml(tif.ome_metadata)

        assert len(ome.images) == 1
        # assert image pixel units are µm
        (pixel_physical_size_unit,) = set(
            [
                ome.images[0].pixels.physical_size_x_unit,
                ome.images[0].pixels.physical_size_y_unit,
                ome.images[0].pixels.physical_size_z_unit,
            ]
        )
        assert (
            pixel_physical_size_unit == UnitsLength.MICROMETER
        ), f"Unit is different than µm, it's {pixel_physical_size_unit.value}"

        image_resampled = _resample(
            image,
            (
                ome.images[0].pixels.physical_size_z / target_pixel_physical_size,
                ome.images[0].pixels.physical_size_y / target_pixel_physical_size,
                ome.images[0].pixels.physical_size_x / target_pixel_physical_size,
            ),
            multichannel=True,
        )

        filename = os.path.join(folder_path, os.path.basename(im_path))
        tifffile.imwrite(
            filename,
            image_resampled,
            metadata={
                "axes": "CZYX",
                "PhysicalSizeX": target_pixel_physical_size,
                "PhysicalSizeXUnit": "µm",
                "PhysicalSizeY": target_pixel_physical_size,
                "PhysicalSizeYUnit": "µm",
                "PhysicalSizeZ": target_pixel_physical_size,
                "PhysicalSizeZUnit": "µm",
            },
        )
        tif.close()


def save_poses(path: str, poses: np.ndarray, names: Optional[list[str]] = None):
    with open(path, "w") as f:
        f.write("name,rot1,rot2,rot3,t1,t2,t3\n")
        for i, p in enumerate(poses):
            pose = list(map(str, p.tolist()))
            name = names[i] if names else str(i)
            f.write(",".join([name] + pose))
            f.write("\n")


def read_poses(path: str, alphabetic_order=True):
    content = csv.reader(open(path, "r").read().split("\n"))
    next(content)
    poses, fnames = [], []
    for row in content:
        if len(row) > 0:
            poses.append(np.array(row[1:], dtype=float))
            fnames.append(row[0])
    if alphabetic_order:
        fnames, poses = zip(*sorted(zip(fnames, poses), key=lambda x: x[0]))
    poses = np.stack(poses)
    return poses, list(fnames)



from typing import Optional, Sequence, Tuple, Union

import cupy as cp
from cucim.skimage.registration import (
    phase_cross_correlation as phase_cross_correlation_cucim,
)
from cupy.typing import DTypeLike
from cupy.typing import NDArray as CPArray
from cupyx.scipy.ndimage import affine_transform as affine_transform_cupy
from cupyx.scipy.ndimage import fourier_shift as fourier_shift_cupy


def affine_transform_batched_multichannel_cupy(
    input: CPArray,
    matrix: CPArray,
    offset: Union[float, Tuple[float], CPArray] = 0.0,
    output_shape: Optional[Tuple[int]] = None,
    output: Optional[Union[CPArray, DTypeLike]] = None,
    order: int = 1,
    mode: str = "constant",
    cval: float = 0.0,
    prefilter: bool = True,
) -> CPArray:
    N, C, *image_shape = input.shape
    if output_shape is None:
        output_shape = tuple(image_shape)
    return_none = False
    if output is None:
        output = cp.empty((N, C) + output_shape, dtype=input.dtype)
    elif type(output) is type:
        output = cp.empty((N, C) + output_shape, dtype=output)
    else:
        return_none = True
    if isinstance(offset, float) or isinstance(offset, tuple):

        def offset_gen(_):
            return offset

    else:
        assert type(offset) is cp.ndarray

        def offset_gen(i):
            return offset[i]

    for i in range(N):
        for j in range(C):
            affine_transform_cupy(
                input[i, j],
                matrix[i],
                offset_gen(i),
                output_shape,
                output[i, j],
                order,
                mode,
                cval,
                prefilter,
                texture_memory=False,
            )

    if return_none:
        return

    return output


def fourier_shift_broadcasted_cupy(
    input: CPArray,
    shift: Union[float, Sequence[float]],
    n: int = -1,
    axis: int = -1,
    output: Optional[CPArray] = None,
):
    shift = cp.asarray(shift)
    if shift.ndim == 0:
        shift = cp.asarray([shift] * input.ndim)
    nb_spatial_dims = shift.shape[-1]
    broadcasted_shape = cp.broadcast_shapes(
        input.shape[:-nb_spatial_dims], shift.shape[:-1]
    )
    image_shape = input.shape[-nb_spatial_dims:]
    input = cp.broadcast_to(input, broadcasted_shape + image_shape)
    shift = cp.broadcast_to(shift, broadcasted_shape + (nb_spatial_dims,))
    output = cp.empty(broadcasted_shape + image_shape, dtype=input.dtype)
    for index in cp.ndindex(broadcasted_shape):
        fourier_shift_cupy(
            input[index].copy(),
            shift[index],
            n,
            axis,
            output[index],
        )
    return output


def phase_cross_correlation_broadcasted_cucim(
    reference_image: CPArray,
    moving_image: CPArray,
    *,
    upsample_factor: int = 1,
    space: str = "real",
    disambiguate: bool = False,
    reference_mask: Optional[CPArray] = None,
    moving_mask: Optional[CPArray] = None,
    overlap_ratio: float = 0.3,
    normalization: str = "phase",
    nb_spatial_dims: Optional[int] = None,
):
    if nb_spatial_dims is None:
        shifts, error, phasediff = phase_cross_correlation_cucim(
            reference_image,
            moving_image,
            upsample_factor=upsample_factor,
            space=space,
            disambiguate=disambiguate,
            return_error="always",
            reference_mask=reference_mask,
            moving_mask=moving_mask,
            overlap_ratio=overlap_ratio,
            normalization=normalization,
        )
        shifts = tuple([cp.array(s, dtype=float) for s in shifts])
        return shifts, error, phasediff

    broadcast = cp.broadcast(reference_image, moving_image)
    reference_image, moving_image = cp.broadcast_arrays(reference_image, moving_image)
    other_shape = broadcast.shape[:-nb_spatial_dims]
    shifts = [cp.empty(other_shape) for _ in range(nb_spatial_dims)]
    errors = cp.empty(other_shape)
    phasediffs = cp.empty(other_shape)
    for index in cp.ndindex(other_shape):
        ref_im, moving_im = reference_image[index], moving_image[index]
        s, e, p = phase_cross_correlation_cucim(
            ref_im,
            moving_im,
            upsample_factor=upsample_factor,
            space=space,
            disambiguate=disambiguate,
            return_error="always",
            reference_mask=reference_mask,
            moving_mask=moving_mask,
            overlap_ratio=overlap_ratio,
            normalization=normalization,
        )
        for i in range(nb_spatial_dims):
            shifts[i][index] = s[i]
        errors[index] = e
        phasediffs[index] = p
    return tuple(shifts), errors, phasediffs



from pathlib import Path

from spfluo import data
from spfluo.ab_initio_reconstruction.__main__ import create_parser, main


def test_main_ab_initio(tmpdir):
    generated_root_dir = data.generated_anisotropic()["rootdir"]
    parser = create_parser()
    tmpdir = Path(tmpdir)
    args = parser.parse_args(
        [
            "--particles_dir",
            str(generated_root_dir / "particles"),
            "--psf_path",
            str(generated_root_dir / "psf.tiff"),
            "--output_dir",
            str(tmpdir),
            "--N_iter_max",
            str(1),
            "--N_axes",
            str(1),
            "--N_rot",
            str(1),
        ]
    )
    main(args)
    reconstruction_path = tmpdir / "final_recons.ome.tiff"
    poses_path = tmpdir / "poses.csv"
    assert reconstruction_path.exists()
    assert poses_path.exists()



import logging
import pathlib
from datetime import datetime
from typing import Any

import numpy as np
import tifffile

DEBUG_DIR = pathlib.Path("spfluo_debug")
DEBUG_DIR_REFINEMENT = DEBUG_DIR / "refinement"


def arg_pretty_repr(arg):
    if hasattr(arg, "shape"):
        return array_pretty_repr(arg)
    else:
        return repr(arg)


def array_pretty_repr(arr):
    return f"({type(arr)}, {arr.shape}, {arr.dtype})"


def create_debug_directories():
    global DEBUG_DIR, DEBUG_DIR_REFINEMENT
    if logging.getLogger("spfluo").isEnabledFor(logging.DEBUG):
        DEBUG_DIR.mkdir(exist_ok=True)
    if logging.getLogger("spfluo.refinement").isEnabledFor(logging.DEBUG):
        DEBUG_DIR_REFINEMENT.mkdir(parents=True, exist_ok=True)


def save_image(
    image: np.ndarray,
    directory: pathlib.Path,
    func: Any,
    *args: str,
    sequence=False,
    multichannel=False,
) -> str:
    create_debug_directories()
    ts = f"{datetime.now().timestamp():.3f}"
    names = "_".join(args)
    path = str(directory / (ts + "_" + func.__name__ + "_" + names)) + ".ome.tiff"
    axes = "ZYX"
    if multichannel:
        axes = "CZYX"
    if sequence:
        axes = "T" + axes
    assert image.ndim == len(axes), f"len({image.shape=} doesnt match {axes=}"
    tifffile.imwrite(path, image, metadata={"axes": axes})
    return path



from __future__ import annotations

import itertools
import logging
import math
from typing import Generator, Tuple

import numpy as np

logger = logging.getLogger("spfluo.utils.memory")


def _split_batch(
    max_batch: Tuple[int | None],
    shape: Tuple[int],
    offset: Tuple[int] = None,
) -> Generator[int | Tuple[int], None, None]:
    if type(max_batch) is tuple:
        assert len(max_batch) == len(shape)
        max_batch = np.array(
            [mb if mb is not None else shape[i] for i, mb in enumerate(max_batch)]
        )
    else:
        max_batch = np.array([max_batch], dtype=int)

    max_batch = np.maximum(max_batch, 1)
    batch_size = np.maximum(2 ** (np.floor(np.log2(max_batch))), 1)
    batch_size = batch_size.astype(int)
    (times, remain_shape) = np.divmod(shape, batch_size)

    if offset is None:
        offset = np.zeros((batch_size.shape[0]), dtype=int)
    ranges = [range(t) for t in times]
    for x in itertools.product(*ranges):
        start = offset + batch_size * np.array(x)
        end = start + batch_size
        out = list(zip(start, end))
        if len(out) == 1:
            logger.debug(
                f"Minibatch summary: minibatch={out[0]}, {max_batch=}, {shape=}"
            )
            yield out[0]
        else:
            logger.debug(f"Minibatch summary: minibatch={out}, {max_batch=}, {shape=}")
            yield out

    if remain_shape.sum() > 0:
        rectangle = times * batch_size
        dims = remain_shape.nonzero()[0]
        if isinstance(shape, int):
            shape = np.array([shape])
        for i in range(1, len(dims) + 1):
            for combination in itertools.combinations(dims, i):
                combination = np.asarray(combination)
                new_shape = rectangle.copy()
                new_shape[combination] = (
                    np.asarray(shape)[combination] - rectangle[combination]
                )
                new_offset = offset.copy()
                new_offset[combination] += rectangle[combination]
                for o in _split_batch(
                    tuple(np.minimum(max_batch, new_shape)),
                    tuple(new_shape),
                    new_offset,
                ):
                    logger.debug(
                        f"Minibatch summary: minibatch={o}, {max_batch=}, {shape=}"
                    )
                    yield o


def find_minibatch(shape: tuple[int], max_size: int) -> tuple[int]:
    if len(shape) == 0:
        return ()
    cumproduct = np.cumprod(np.flip(np.asarray(shape)))
    i = np.sum(cumproduct <= max_size)
    if i == 0:
        return (max_size,)
    else:
        return (
            find_minibatch(shape[:-i], math.floor(max_size / cumproduct[i - 1]))
            + shape[-i:]
        )


def split_batch(
    shape: Tuple[int],
    max_batch: int | None = None,
    use_memory: bool = False,
) -> Generator[int | Tuple[int], None, None]:
    if use_memory:
        raise NotImplementedError()

    if max_batch is None:
        yield from _split_batch(tuple([None for _ in range(len(shape))]), shape)
        return

    minibatch = find_minibatch(shape, max_batch)
    yield from _split_batch(
        tuple([0 for i in range(len(shape) - len(minibatch))]) + minibatch, shape
    )


if __name__ == "__main__":
    for o in split_batch((8, None, 10), (100, 120, 17)):
        print(o)



from typing import TYPE_CHECKING

import numpy
from hypothesis import assume
from hypothesis import strategies as st
from scipy.spatial.transform import Rotation as R

import spfluo
from spfluo.utils.array import array_namespace, get_cupy, get_torch, to_numpy
from spfluo.utils.array import numpy as np
from spfluo.utils.volume import phase_cross_correlation

if TYPE_CHECKING:
    from spfluo.utils.array import Array


testing_libs = [(np, "cpu")]
ids = ["numpy"]

try:
    import array_api_strict as xp
    from array_api_strict._array_object import CPU_DEVICE

    testing_libs.append((xp, CPU_DEVICE))
    ids.append("array-api-strict")
except ModuleNotFoundError:
    pass


if spfluo.has_cupy():
    cupy = get_cupy()
    testing_libs.append((cupy, cupy.cuda.Device(0)))
    ids.append("cupy")

if spfluo.has_torch():
    torch = get_torch()
    testing_libs.append((torch, "cpu"))
    ids.append("torch-cpu")
    if spfluo.has_torch_cuda():
        testing_libs.append((torch, "cuda"))
        ids.append("torch-cuda")


@st.composite
def random_pose(
    draw,
    quaternions=st.tuples(
        st.floats(0, 1, allow_infinity=False, allow_nan=False),
        st.floats(0, 1, allow_infinity=False, allow_nan=False),
        st.floats(0, 1, allow_infinity=False, allow_nan=False),
        st.floats(0, 1, allow_infinity=False, allow_nan=False),
    ),
    translation=st.tuples(
        st.floats(0, 1, allow_infinity=False, allow_nan=False),
        st.floats(0, 1, allow_infinity=False, allow_nan=False),
        st.floats(0, 1, allow_infinity=False, allow_nan=False),
    ),
    translation_magnitude=st.floats(0, 1, allow_infinity=False, allow_nan=False),
):
    quaternions = draw(quaternions)
    translation = draw(translation)
    translation_magnitude = draw(translation_magnitude)

    assume(np.linalg.norm(quaternions) > 0)
    assume(not np.isinf(np.linalg.norm(quaternions)))
    assume(np.linalg.norm(translation) > 0)
    assume(not np.isinf(np.linalg.norm(translation)))

    euler = R.from_quat(quaternions).as_euler("XZX", degrees=True)
    translation /= np.linalg.norm(translation)
    translation *= translation_magnitude
    pose = np.concatenate((euler, translation))
    return pose


def assert_volumes_aligned(
    vol1: "Array",
    vol2: "Array",
    atol: float = 0.1,
    nb_spatial_dims: int = 3,
    multichannel: bool = False,
):
    xp = array_namespace(vol1, vol2)
    (dz, dy, dx), _, _ = phase_cross_correlation(
        xp.astype(vol1, vol2.dtype),
        vol2,
        upsample_factor=10,
        disambiguate=False,
        normalization=None,
        nb_spatial_dims=nb_spatial_dims,
        multichannel=multichannel,
    )
    n = (dz**2 + dy**2 + dx**2) ** 0.5
    assert xp.all(n <= atol), f"{xp.max(n)} > {atol}"


def assert_allclose(actual: "Array", desired: "Array", rtol=1e-7, atol=0):
    numpy.testing.assert_allclose(
        to_numpy(actual), to_numpy(desired), rtol=rtol, atol=atol
    )



import argparse
import logging
import os

from spfluo.utils.log import base_parser, set_logging_level

from .loading import resample, resize
from .rotate_symmetry_axis import main as rotate_symmetry_axis_main

utils_logger = logging.getLogger("spfluo.utils")


def create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser("Utils functions", parents=[base_parser])
    parser.add_argument("-f", "--function", type=str)

    # common args
    parser.add_argument(
        "-i", "--input", type=str, help="The image(s) to process", nargs="+"
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="The path to the output image/directory",
        default=None,
    )

    # resize args
    parser.add_argument("--size", type=float, help="Size in µm")

    # resample args
    parser.add_argument(
        "--target_pixel_size", type=float, help="Target pixel size in µm", default=None
    )

    # rotate_symmetry_axis args
    parser.add_argument("--symmetry", type=int, help="symmetry degree of the particle")
    parser.add_argument(
        "--convention", type=str, help="scipy rotation convention", default="XZX"
    )
    parser.add_argument("--poses", type=str, help="path to poses", default=None)
    parser.add_argument(
        "--rotated-volume", type=str, help="path to the rotated volume", default=None
    )
    parser.add_argument(
        "--rotated-poses", type=str, help="path to the rotated poses", default=None
    )
    parser.add_argument(
        "--threshold",
        type=float,
        help="threshold in percentage of the max",
        default=0.5,
    )

    return parser


def main(args: argparse.Namespace) -> None:
    if args.input is None:
        parser.print_help()
        return
    image_paths = list(map(os.path.abspath, args.input))
    output_path = os.path.abspath(args.output) if args.output else None
    utils_logger.info("Function : " + args.function)
    utils_logger.debug("Images :" + str(image_paths))
    if args.function == "resize":
        resize(image_paths, args.size, output_path)
    if args.function == "resample":
        resample(image_paths, output_path, args.target_pixel_size)
    if args.function == "rotate_symmetry_axis":
        assert len(args.input) == 1
        rotate_symmetry_axis_main(
            args.input[0],
            args.symmetry,
            args.convention,
            threshold=args.threshold,
            output_volume_path=args.rotated_volume,
            poses_path=args.poses,
            output_poses_path=args.rotated_poses,
        )


if __name__ == "__main__":
    parser = create_parser()
    args = parser.parse_args()
    set_logging_level(args)
    main(args)



"""
Code from https://github.com/wolny/pytorch-3dunet/blob/master/pytorch3dunet/unet3d/losses.py
"""

import torch
import torch.nn.functional as F
from torch import nn as nn
from torch.autograd import Variable
from torch.nn import L1Loss, MSELoss, SmoothL1Loss


def expand_as_one_hot(input, C, ignore_index=None):
    """
    Converts NxSPATIAL label image to NxCxSPATIAL, where each label gets converted to its corresponding one-hot vector.
    It is assumed that the batch dimension is present.
    Args:
        input (torch.Tensor): 3D/4D input image
        C (int): number of channels/labels
        ignore_index (int): ignore index to be kept during the expansion
    Returns:
        4D/5D output torch.Tensor (NxCxSPATIAL)
    """
    assert input.dim() == 4

    # expand the input tensor to Nx1xSPATIAL before scattering
    input = input.unsqueeze(1)
    # create output tensor shape (NxCxSPATIAL)
    shape = list(input.size())
    shape[1] = C

    if ignore_index is not None:
        # create ignore_index mask for the result
        mask = input.expand(shape) == ignore_index
        # clone the src tensor and zero out ignore_index in the input
        input = input.clone()
        input[input == ignore_index] = 0
        # scatter to get the one-hot tensor
        result = torch.zeros(shape).to(input.device).scatter_(1, input, 1)
        # bring back the ignore_index in the result
        result[mask] = ignore_index
        return result
    else:
        # scatter to get the one-hot tensor
        return torch.zeros(shape).to(input.device).scatter_(1, input, 1)


def compute_per_channel_dice(input, target, epsilon=1e-6, weight=None):
    """
    Computes DiceCoefficient as defined in https://arxiv.org/abs/1606.04797 given  a multi channel input and target.
    Assumes the input is a normalized probability, e.g. a result of Sigmoid or Softmax function.

    Args:
         input (torch.Tensor): NxCxSpatial input tensor
         target (torch.Tensor): NxCxSpatial target tensor
         epsilon (float): prevents division by zero
         weight (torch.Tensor): Cx1 tensor of weight per channel/class
    """

    # input and target shapes must match
    assert (
        input.size() == target.size()
    ), "'input' and 'target' must have the same shape"

    input = flatten(input)
    target = flatten(target)
    target = target.float()

    # compute per channel Dice Coefficient
    intersect = (input * target).sum(-1)
    if weight is not None:
        intersect = weight * intersect

    # here we can use standard dice (input + target).sum(-1) or extension (see V-Net) (input^2 + target^2).sum(-1)
    denominator = (input * input).sum(-1) + (target * target).sum(-1)
    return 2 * (intersect / denominator.clamp(min=epsilon))


class _MaskingLossWrapper(nn.Module):
    """
    Loss wrapper which prevents the gradient of the loss to be computed where target is equal to `ignore_index`.
    """

    def __init__(self, loss, ignore_index):
        super(_MaskingLossWrapper, self).__init__()
        assert ignore_index is not None, "ignore_index cannot be None"
        self.loss = loss
        self.ignore_index = ignore_index

    def forward(self, input, target):
        mask = target.clone().ne_(self.ignore_index)
        mask.requires_grad = False

        # mask out input/target so that the gradient is zero where on the mask
        input = input * mask
        target = target * mask

        # forward masked input and target to the loss
        return self.loss(input, target)


class SkipLastTargetChannelWrapper(nn.Module):
    """
    Loss wrapper which removes additional target channel
    """

    def __init__(self, loss, squeeze_channel=False):
        super(SkipLastTargetChannelWrapper, self).__init__()
        self.loss = loss
        self.squeeze_channel = squeeze_channel

    def forward(self, input, target):
        assert (
            target.size(1) > 1
        ), "Target tensor has a singleton channel dimension, cannot remove channel"

        # skips last target channel if needed
        target = target[:, :-1, ...]

        if self.squeeze_channel:
            # squeeze channel dimension if singleton
            target = torch.squeeze(target, dim=1)
        return self.loss(input, target)


class _AbstractDiceLoss(nn.Module):
    """
    Base class for different implementations of Dice loss.
    """

    def __init__(self, weight=None, normalization="sigmoid"):
        super(_AbstractDiceLoss, self).__init__()
        self.register_buffer("weight", weight)
        # The output from the network during training is assumed to be un-normalized probabilities and we would
        # like to normalize the logits. Since Dice (or soft Dice in this case) is usually used for binary data,
        # normalizing the channels with Sigmoid is the default choice even for multi-class segmentation problems.
        # However if one would like to apply Softmax in order to get the proper probability distribution from the
        # output, just specify `normalization=Softmax`
        assert normalization in ["sigmoid", "softmax", "none"]
        if normalization == "sigmoid":
            self.normalization = nn.Sigmoid()
        elif normalization == "softmax":
            self.normalization = nn.Softmax(dim=1)
        else:
            self.normalization = lambda x: x

    def dice(self, input, target, weight):
        # actual Dice score computation; to be implemented by the subclass
        raise NotImplementedError

    def forward(self, input, target):
        # get probabilities from logits
        input = self.normalization(input)

        # compute per channel Dice coefficient
        per_channel_dice = self.dice(input, target, weight=self.weight)

        # average Dice score across all channels/classes
        return 1.0 - torch.mean(per_channel_dice)


class DiceLoss(_AbstractDiceLoss):
    """Computes Dice Loss according to https://arxiv.org/abs/1606.04797.
    For multi-class segmentation `weight` parameter can be used to assign different weights per class.
    The input to the loss function is assumed to be a logit and will be normalized by the Sigmoid function.
    """

    def __init__(self, weight=None, normalization="sigmoid"):
        super().__init__(weight, normalization)

    def dice(self, input, target, weight):
        return compute_per_channel_dice(input, target, weight=self.weight)


class GeneralizedDiceLoss(_AbstractDiceLoss):
    """Computes Generalized Dice Loss (GDL) as described in https://arxiv.org/pdf/1707.03237.pdf."""

    def __init__(self, normalization="sigmoid", epsilon=1e-6):
        super().__init__(weight=None, normalization=normalization)
        self.epsilon = epsilon

    def dice(self, input, target, weight):
        assert (
            input.size() == target.size()
        ), "'input' and 'target' must have the same shape"

        input = flatten(input)
        target = flatten(target)
        target = target.float()

        if input.size(0) == 1:
            # for GDL to make sense we need at least 2 channels (see https://arxiv.org/pdf/1707.03237.pdf)
            # put foreground and background voxels in separate channels
            input = torch.cat((input, 1 - input), dim=0)
            target = torch.cat((target, 1 - target), dim=0)

        # GDL weighting: the contribution of each label is corrected by the inverse of its volume
        w_l = target.sum(-1)
        w_l = 1 / (w_l * w_l).clamp(min=self.epsilon)
        w_l.requires_grad = False

        intersect = (input * target).sum(-1)
        intersect = intersect * w_l

        denominator = (input + target).sum(-1)
        denominator = (denominator * w_l).clamp(min=self.epsilon)

        return 2 * (intersect.sum() / denominator.sum())


class BCEDiceLoss(nn.Module):
    """Linear combination of BCE and Dice losses"""

    def __init__(self, alpha, beta):
        super(BCEDiceLoss, self).__init__()
        self.alpha = alpha
        self.bce = nn.BCEWithLogitsLoss()
        self.beta = beta
        self.dice = DiceLoss()

    def forward(self, input, target):
        return self.alpha * self.bce(input, target) + self.beta * self.dice(
            input, target
        )


class WeightedCrossEntropyLoss(nn.Module):
    """WeightedCrossEntropyLoss (WCE) as described in https://arxiv.org/pdf/1707.03237.pdf"""

    def __init__(self, ignore_index=-1):
        super(WeightedCrossEntropyLoss, self).__init__()
        self.ignore_index = ignore_index

    def forward(self, input, target):
        weight = self._class_weights(input)
        return F.cross_entropy(
            input, target, weight=weight, ignore_index=self.ignore_index
        )

    @staticmethod
    def _class_weights(input):
        # normalize the input first
        input = F.softmax(input, dim=1)
        flattened = flatten(input)
        nominator = (1.0 - flattened).sum(-1)
        denominator = flattened.sum(-1)
        class_weights = Variable(nominator / denominator, requires_grad=False)
        return class_weights


class PixelWiseCrossEntropyLoss(nn.Module):
    def __init__(self, class_weights=None, ignore_index=None):
        super(PixelWiseCrossEntropyLoss, self).__init__()
        self.register_buffer("class_weights", class_weights)
        self.ignore_index = ignore_index
        self.log_softmax = nn.LogSoftmax(dim=1)

    def forward(self, input, target, weights):
        assert target.size() == weights.size()
        # normalize the input
        log_probabilities = self.log_softmax(input)
        # standard CrossEntropyLoss requires the target to be (NxDxHxW), so we need to expand it to (NxCxDxHxW)
        target = expand_as_one_hot(
            target, C=input.size()[1], ignore_index=self.ignore_index
        )
        # expand weights
        weights = weights.unsqueeze(1)
        weights = weights.expand_as(input)

        # create default class_weights if None
        if self.class_weights is None:
            class_weights = torch.ones(input.size()[1]).float().to(input.device)
        else:
            class_weights = self.class_weights

        # resize class_weights to be broadcastable into the weights
        class_weights = class_weights.view(1, -1, 1, 1, 1)

        # multiply weights tensor by class weights
        weights = class_weights * weights

        # compute the losses
        result = -weights * target * log_probabilities
        # average the losses
        return result.mean()


class WeightedSmoothL1Loss(nn.SmoothL1Loss):
    def __init__(self, threshold, initial_weight, apply_below_threshold=True):
        super().__init__(reduction="none")
        self.threshold = threshold
        self.apply_below_threshold = apply_below_threshold
        self.weight = initial_weight

    def forward(self, input, target):
        l1 = super().forward(input, target)

        if self.apply_below_threshold:
            mask = target < self.threshold
        else:
            mask = target >= self.threshold

        l1[mask] = l1[mask] * self.weight

        return l1.mean()


class iaawnLoss(nn.Module):
    """intersections are all we need Loss"""

    def __init__(self, normalization="sigmoid", alpha=1):
        super().__init__()
        self.alpha = alpha
        if normalization == "sigmoid":
            self.normalization = nn.Sigmoid()
        else:
            self.normalization = lambda x: x

    def forward(self, input, target, epsilon=1e-6):
        input = self.normalization(input)

        input = flatten(input)
        target = flatten(target)

        mask1 = (target == 1).float()
        mask2 = (target > 1).float()

        den1 = mask1.sum(-1).clamp(epsilon, None)
        den2 = mask2.sum(-1).clamp(epsilon, None)

        loss = (input * mask1).sum(-1) / den1 - self.alpha * (input * mask2).sum(
            -1
        ) / den2
        return -torch.mean(loss)


class iaawnMetric(nn.Module):
    """intersections are all we need Metric"""

    def __init__(self, normalization="sigmoid", alpha=1):
        super().__init__()
        self.iaawnLoss = iaawnLoss(normalization="sigmoid", alpha=1)

    def forward(self, input, target, epsilon=1e-6):
        loss = self.iaawnLoss(input, target, epsilon=epsilon)
        return (-loss + 1) / 2


class iaawnLossDenis(nn.Module):
    """intersections are all we need Loss"""

    def __init__(self, normalization="sigmoid", alpha=1, beta=0.1):
        super().__init__()
        self.alpha = alpha
        self.beta = beta
        if normalization == "sigmoid":
            self.normalization = nn.Sigmoid()
        else:
            self.normalization = lambda x: x

    def forward(self, input, target, epsilon=1e-6):
        input = self.normalization(input)

        input = flatten(input)
        target = flatten(target)

        mask1 = (target == 1).float()
        mask2 = (target > 1).float()

        den1 = mask1.sum(-1)
        den2 = mask2.sum(-1)
        no_mask = torch.logical_and(den1 < epsilon, den2 < epsilon)
        den1, den2 = den1.clamp(epsilon, None), den2.clamp(epsilon, None)

        third_term = (
            torch.logical_not(no_mask) * self.beta * input.sum(-1) / (den1 + den2)
        )

        loss = (
            (input * mask1).sum(-1) / den1
            - self.alpha * (input * mask2).sum(-1) / den2
            + third_term
        )
        return -torch.mean(loss)


def flatten(tensor):
    """Flattens a given tensor such that the channel axis is first.
    The shapes are transformed as follows:
       (N, C, D, H, W) -> (C, N * D * H * W)
    """
    # number of channels
    C = tensor.size(1)
    # new axis order
    axis_order = (1, 0) + tuple(range(2, tensor.dim()))
    # Transpose: (N, C, D, H, W) -> (C, N, D, H, W)
    transposed = tensor.permute(axis_order)
    # Flatten: (C, N, D, H, W) -> (C, N * D * H * W)
    return transposed.contiguous().view(C, -1)


def get_loss_criterion(config):
    """
    Returns the loss function based on provided configuration
    :param config: (dict) a top level configuration object containing the 'loss' key
    :return: an instance of the loss function
    """
    assert "loss" in config, "Could not find loss function configuration"
    loss_config = config["loss"]
    name = loss_config.pop("name")

    ignore_index = loss_config.pop("ignore_index", None)
    skip_last_target = loss_config.pop("skip_last_target", False)
    weight = loss_config.pop("weight", None)

    if weight is not None:
        # convert to cuda tensor if necessary
        weight = torch.tensor(weight).to(config["device"])

    pos_weight = loss_config.pop("pos_weight", None)
    if pos_weight is not None:
        # convert to cuda tensor if necessary
        pos_weight = torch.tensor(pos_weight).to(config["device"])

    loss = _create_loss(name, loss_config, weight, ignore_index, pos_weight)

    if not (
        ignore_index is None or name in ["CrossEntropyLoss", "WeightedCrossEntropyLoss"]
    ):
        # use MaskingLossWrapper only for non-cross-entropy losses, since CE losses allow specifying 'ignore_index' directly
        loss = _MaskingLossWrapper(loss, ignore_index)

    if skip_last_target:
        loss = SkipLastTargetChannelWrapper(
            loss, loss_config.get("squeeze_channel", False)
        )

    return loss


#######################################################################################################################


def _create_loss(name, loss_config, weight, ignore_index, pos_weight):
    if name == "BCEWithLogitsLoss":
        return nn.BCEWithLogitsLoss(pos_weight=pos_weight)
    elif name == "BCEDiceLoss":
        alpha = loss_config.get("alphs", 1.0)
        beta = loss_config.get("beta", 1.0)
        return BCEDiceLoss(alpha, beta)
    elif name == "CrossEntropyLoss":
        if ignore_index is None:
            ignore_index = (
                -100
            )  # use the default 'ignore_index' as defined in the CrossEntropyLoss
        return nn.CrossEntropyLoss(weight=weight, ignore_index=ignore_index)
    elif name == "WeightedCrossEntropyLoss":
        if ignore_index is None:
            ignore_index = (
                -100
            )  # use the default 'ignore_index' as defined in the CrossEntropyLoss
        return WeightedCrossEntropyLoss(ignore_index=ignore_index)
    elif name == "PixelWiseCrossEntropyLoss":
        return PixelWiseCrossEntropyLoss(
            class_weights=weight, ignore_index=ignore_index
        )
    elif name == "GeneralizedDiceLoss":
        normalization = loss_config.get("normalization", "sigmoid")
        return GeneralizedDiceLoss(normalization=normalization)
    elif name == "DiceLoss":
        normalization = loss_config.get("normalization", "sigmoid")
        return DiceLoss(weight=weight, normalization=normalization)
    elif name == "MSELoss":
        return MSELoss()
    elif name == "SmoothL1Loss":
        return SmoothL1Loss()
    elif name == "L1Loss":
        return L1Loss()
    elif name == "WeightedSmoothL1Loss":
        return WeightedSmoothL1Loss(
            threshold=loss_config["threshold"],
            initial_weight=loss_config["initial_weight"],
            apply_below_threshold=loss_config.get("apply_below_threshold", True),
        )
    else:
        raise RuntimeError(f"Unsupported loss function: '{name}'")



# /// script
# dependencies = [
#   "plotly",
#   "numpy",
#   "torch",
#
# ]
# ///

from __future__ import annotations

import logging
from typing import Sequence, Union

import plotly.express as px
import plotly.graph_objects as go
from numpy.typing import NDArray
from torch import Tensor


def get_all_named_cmaps() -> list[str]:
    named_colorscales = px.colors.named_colorscales()
    base_cmaps = ["cividis", "sunset", "turbo", "thermal"]
    base_cmaps_idx = [named_colorscales.index(x) for x in base_cmaps]
    for idx in base_cmaps_idx:
        named_colorscales.pop(idx)
    named_colorscales = base_cmaps + named_colorscales
    return named_colorscales


def interactive_plot(
    data: Union[NDArray, Sequence[NDArray]],
    labels: Union[str, Sequence[str]] = None,
    point_size: int = 3,
    opacity: float = 0.8,
    colorbar: bool = False,
    color: NDArray = None,
    color_range: tuple[int] = None,
    cmap: str = None,
    constraint_x: bool = False,
    constraint_y: bool = False,
    constraint_z: bool = False,
    return_fig: bool = False,
    width: int = 500,
    height: int = 500,
    title: str = None,
):
    """Interactive plot of point cloud(s) based on Plotly. Can display N pointcloud(s).

    Args:
        pointcloud (Union[np.ndarray, tuple[np.ndarray]]): If a list or tuple is passed,
            each element will be displayed with its own name, colormap, and button to
            toggle visibility.
        label (str, optional): This will be the title of the plot. Is is called label
            because it was intended to be used within a classification setup.
            Defaults to None.
        point_size (int, optional): Display size of one point (x, y, z). Defaults to 1.
        color (np.ndarray, optional): Color of each points. MUST be a sequence of length
            equals to the number of points.
            If None, the z coordinates will be used to color points.
            Defaults to None.
        constraint_x (bool, optional): Rescale x within [-1 ,1]. Defaults to False.
        constraint_y (bool, optional): Rescale y within [-1 ,1]. Defaults to False.
        constraint_z (bool, optional): Rescale z within [-1 ,1]. Defaults to False.

    Raises:
        ValueError: If too few or too many pointclouds are passed to the function.
    """
    if not isinstance(data, (list, tuple)):
        data = [data]
    if isinstance(data, tuple):
        data = list(data)
    # move to cpu if required; raises warning
    moved_to_cpu = False
    for i, x in enumerate(data):
        if isinstance(x, Tensor) and x.is_cuda:
            data[i] = x.cpu()
            moved_to_cpu = True
    if moved_to_cpu:
        logging.basicConfig(format="[%(levelname)s] %(message)s")
        logging.warning(
            (
                "CUDA Tensors were passed in: "
                "I had to move them to the cpu in order to display them."
            )
        )
    N = len(data)
    if labels is None:
        labels = [f"pointcloud {i + 1}" for i in range(N)]
    if not isinstance(labels, (list, tuple)):
        labels = (labels,)
    if labels is not None and not len(data) == len(labels):
        raise ValueError(f"You gave {len(data)} pointclouds but {len(labels)} labels.")
    all_cmaps = get_all_named_cmaps()
    if isinstance(cmap, str):
        cmaps = N * [cmap]
    elif not (isinstance(cmap, (tuple, list)) and len(cmap) == len(data)):
        cmaps = all_cmaps[:N]
    else:
        cmaps = cmap
    traces = list()
    for pointcloud, label, cmap in zip(data, labels, cmaps):
        x, y, z = pointcloud[:, 0], pointcloud[:, 1], pointcloud[:, 2]
        c = color if color is not None else z
        marker_kwargs = dict(size=point_size, opacity=opacity, color=c, colorscale=cmap)
        if color_range is not None:
            marker_kwargs["cmin"] = color_range[0]
            marker_kwargs["cmax"] = color_range[1]
        if colorbar:
            marker_kwargs["colorbar"] = dict(thickness=20)
        scatter_kwargs = dict(
            visible=True, mode="markers", name=label, marker=marker_kwargs
        )
        traces.append(go.Scatter3d(x=x, y=y, z=z, **scatter_kwargs))
    layout = dict(
        width=width,
        height=height,
        xaxis=dict(range=[0, 1]),
        yaxis=dict(range=[0, 1]),
        margin=dict(t=50),
    )
    layout["scene"] = dict()
    if constraint_x:
        layout["scene"]["xaxis"] = dict(nticks=4, range=[-1, 1])
    if constraint_y:
        layout["scene"]["yaxis"] = dict(nticks=4, range=[-1, 1])
    if constraint_z:
        layout["scene"]["zaxis"] = dict(nticks=4, range=[-1, 1])
    if title is not None:
        layout["title"] = title
    fig = go.Figure(data=traces, layout=layout)
    if return_fig:
        return fig
    fig.show()



import csv
import os
import pickle
from pathlib import Path
from typing import List, Optional, Tuple

import numpy as np
import numpy.random as R
import scipy.ndimage as ndii
import tifffile
import torch
from pyfigtree import figtree
from scipy.ndimage import gaussian_filter
from scipy.stats import truncnorm
from skimage.util import random_noise
from tqdm import tqdm

from spfluo.ab_initio_reconstruction.volume_representation.gaussian_mixture_representation.GMM_grid_evaluation import (
    make_grid,
    nd_gaussian,
)

from . import functional as F
from .config import DataGenerationConfig
from .config import Outliers as OutliersConfig


class DataGenerator:
    # +---------------------------------------------------------------------------------------+ #
    # |                                        INIT                                           | #
    # +---------------------------------------------------------------------------------------+ #

    def __init__(self, config: DataGenerationConfig) -> None:
        self.config = config
        self.device = self.__set_device()
        self.dtype = self.config.dtype
        self.template_pointcloud = self.__load_pointcloud_tensor()
        x_min, x_max, y_min, y_max, z_min, z_max = F.get_FOV(self.template_pointcloud)
        max_delta = max(x_max - x_min, y_max - y_min, z_max - z_min)
        self.step = float(max_delta / self.config.voxelisation.max_particle_dim)
        self.output_image = np.zeros(self.config.target_shape)

        # Particles FOV
        image_shape = self.config.voxelisation.image_shape
        eps = 1e-12 / 2
        self.common_fov = (
            -(self.step * image_shape - eps) / 2,
            (self.step * image_shape - eps) / 2,
        ) * 3

        self.create_psf()

    def __load_pointcloud_tensor(self) -> torch.Tensor:
        template_point_cloud = np.loadtxt(
            self.config.io.point_cloud_path, delimiter=","
        )
        pointcloud = torch.as_tensor(template_point_cloud, dtype=torch.float32).to(
            self.device
        )
        return (pointcloud - pointcloud.mean(dim=0, keepdim=True))[:, [2, 1, 0]]

    def __set_device(self):
        if self.config.disable_cuda:
            return torch.device("cpu")
        return torch.device("cuda" if torch.cuda.is_available() else "cpu")

    # +---------------------------------------------------------------------------------------+ #
    # |                                  GENERATE CENTERS                                     | #
    # +---------------------------------------------------------------------------------------+ #

    def generate_centers_one_cluster(
        self, remaining_particles: int, deterministic: bool
    ):
        shape = self.config.target_shape
        radius_xy_min, radius_xy_max = self.config.voxelisation.cluster_range_xy
        radius_z_min, radius_z_max = self.config.voxelisation.cluster_range_z
        radius_xy = R.randint(radius_xy_min, high=radius_xy_max, size=2)
        radius_z = R.randint(radius_z_min, high=radius_z_max)
        margin = radius_xy + self.config.voxelisation.max_particle_dim
        center_xy = R.randint(margin, high=np.array(shape[1:]) - margin)
        margin = radius_z + self.config.voxelisation.max_particle_dim
        center_z = R.randint(margin, high=np.array(shape[0]) - margin)
        center = (center_z, *center_xy)
        radius = (radius_z, *radius_xy)
        if deterministic:
            nb_particles = remaining_particles
        else:
            nb_min, nb_max = self.config.voxelisation.nb_particles_per_cluster_range
            nb_max = min(nb_max, remaining_particles)
            nb_particles = R.randint(nb_min, high=nb_max + 1)
        margin = np.array(3 * (self.config.voxelisation.max_particle_dim,))
        clip_min, clip_max = (
            (margin - center) / radius,
            (shape - margin - center) / radius,
        )
        truncnorm_kwargs = {"loc": center, "scale": radius, "size": (nb_particles, 3)}
        centers = truncnorm.rvs(clip_min, clip_max, **truncnorm_kwargs)
        return centers.astype(np.uint16)

    """
    def generate_centers(self):
        total_nb_to_generate = self.config.voxelisation.num_particles
        remaining_number_to_generate = total_nb_to_generate
        min_nb_particles_per_cluster = self.config.voxelisation.nb_particles_per_cluster_range[0]
        all_centers = np.ndarray((0, 3), dtype=np.uint16)
        while remaining_number_to_generate > 0:
            deterministic = (remaining_number_to_generate < min_nb_particles_per_cluster)
            args = [remaining_number_to_generate, deterministic]
            centers = self.generate_centers_one_cluster(*args)
            all_centers = np.vstack((all_centers, centers))
            remaining_number_to_generate = total_nb_to_generate - all_centers.shape[0]
        return all_centers
    """

    def generate_centers(self) -> np.ndarray:
        shape = self.config.target_shape
        margin = np.array(3 * (self.config.voxelisation.max_particle_dim,))
        num_to_generate = self.config.voxelisation.num_particles
        centers = np.zeros((1, 3), dtype=np.uint16)
        while len(centers) < num_to_generate + 1:
            center = np.random.randint(margin, shape - margin, size=(1, 3))
            if (
                np.linalg.norm(centers - center, axis=1).min()
                >= 1.5 * self.config.voxelisation.max_particle_dim
            ):
                centers = np.vstack((centers, center))
        return centers[1:]

    # +---------------------------------------------------------------------------------------+ #
    # |                                TILT VIEWS HANDLERS                                    | #
    # +---------------------------------------------------------------------------------------+ #

    @staticmethod
    def sample_tilt_angles(
        num_particles: int, tilt_strategy: int, margin: int = 10
    ) -> List:
        distributions = {}
        distributions["gaussian_top_side_only"] = [
            {"type": R.normal, "kwargs": {"loc": 0, "scale": 10}},
            {"type": R.normal, "kwargs": {"loc": 90, "scale": 10}},
            {"type": R.normal, "kwargs": {"loc": 180, "scale": 10}},
            {"type": R.normal, "kwargs": {"loc": 270, "scale": 10}},
        ]
        distributions["uniform_top_side_only"] = [
            {"type": R.uniform, "kwargs": {"low": 0 - margin, "high": 0 + margin}},
            {"type": R.uniform, "kwargs": {"low": 90 - margin, "high": 90 + margin}},
            {"type": R.uniform, "kwargs": {"low": 180 - margin, "high": 180 + margin}},
            {"type": R.uniform, "kwargs": {"low": 270 - margin, "high": 270 + margin}},
        ]
        distributions["uniform"] = [
            # top and side
            {"type": R.uniform, "kwargs": {"low": 0 - margin, "high": 0 + margin}},
            {"type": R.uniform, "kwargs": {"low": 90 - margin, "high": 90 + margin}},
            {"type": R.uniform, "kwargs": {"low": 180 - margin, "high": 180 + margin}},
            {"type": R.uniform, "kwargs": {"low": 270 - margin, "high": 270 + margin}},
            # other
            {
                "type": R.uniform,
                "kwargs": {"low": 0 + 2 * margin, "high": 90 - 2 * margin},
            },
            {
                "type": R.uniform,
                "kwargs": {"low": 90 + 2 * margin, "high": 180 - 2 * margin},
            },
            {
                "type": R.uniform,
                "kwargs": {"low": 180 + 2 * margin, "high": 270 - 2 * margin},
            },
        ]
        distributions["gaussian"] = [
            # top and side
            {"type": R.normal, "kwargs": {"loc": 0, "scale": margin // 2}},
            {"type": R.normal, "kwargs": {"loc": 90, "scale": margin // 2}},
            {"type": R.normal, "kwargs": {"loc": 180, "scale": margin // 2}},
            {"type": R.normal, "kwargs": {"loc": 270, "scale": margin // 2}},
            # other
            {"type": R.normal, "kwargs": {"loc": 45, "scale": margin}},
            {"type": R.normal, "kwargs": {"loc": 135, "scale": margin}},
            {"type": R.normal, "kwargs": {"loc": 225, "scale": margin}},
        ]
        distributions = distributions[tilt_strategy]
        num_distributions = len(distributions)
        weights = np.ones((num_distributions))
        weights /= weights.sum()
        data = np.zeros((num_particles, num_distributions))
        for idx, distr in enumerate(distributions):
            data[:, idx] = distr["type"](size=(num_particles,), **distr["kwargs"])
        idx_kwargs = {"size": (num_particles,), "p": weights}
        random_idx = R.choice(np.arange(num_distributions), **idx_kwargs)
        return data[np.arange(num_particles), random_idx]

    @staticmethod
    def get_view_from_angles(angles: Tuple[int], margin: int = 10) -> int:
        angle = angles[1]
        around0 = (0 - margin) <= angle <= (0 + margin)
        around180 = (180 - margin) <= angle <= (180 + margin)
        around90 = (90 - margin) <= angle <= (90 + margin)
        around270 = (270 - margin) <= angle <= (270 + margin)
        if around0 or around180:
            return 0  # TOP
        if around90 or around270:
            return 1  # SIDE
        return 2  # OTHER

    def generate_angles(self, tilt_strategy: str, tilt_margin: int) -> np.ndarray:
        num_particles = self.config.voxelisation.num_particles
        tilt_angles = R.uniform(low=0, high=180, size=(num_particles,))
        other_angles = R.uniform(low=0, high=360, size=(num_particles, 2))
        angles = np.vstack((other_angles[:, 0], tilt_angles, other_angles[:, 1])).T
        angles[
            R.uniform(size=(num_particles,)) > self.config.augmentation.rotation_proba
        ] = [0, 0, 0]
        return angles

    def generate_translations(self, max_translation: float) -> np.ndarray:
        num_particles = self.config.voxelisation.num_particles
        return R.uniform(
            low=-max_translation, high=max_translation, size=(num_particles, 3)
        )

    # +---------------------------------------------------------------------------------------+ #
    # |                                 AUGMENT POINTCLOUD                                    | #
    # +---------------------------------------------------------------------------------------+ #

    def augment_pointcloud(
        self,
        shrink_range: Optional[Tuple[float, float]] = None,
        rotation_angles: Tuple[int] = None,
        translation: Tuple[float] = None,
    ) -> np.ndarray:
        cfg = self.config.augmentation
        # 1. Shrink
        a, b = shrink_range if shrink_range else cfg.shrink_range
        factor = (b - a) * R.random_sample() + a
        pointcloud = F.shrink(self.template_pointcloud, factor)
        # 2. Rotate
        pointcloud = F.rotate(pointcloud, rotation_angles, self.device)
        # 3. Translate
        pointcloud = F.translate(pointcloud, translation, self.device)
        # 4. Make holes
        # a, b = cfg.intensity_std_ratio_range
        # std_ratio = (b - a) * R.random_sample() + a
        # pointcloud = F.non_uniform_density(
        #    pointcloud,
        #    cfg.intensity_nb_holes_min,
        #    cfg.intensity_nb_holes_max,
        #    cfg.intensity_mean_ratio,
        #    std_ratio,
        #    self.device
        # )
        return pointcloud.cpu().numpy()

    # +---------------------------------------------------------------------------------------+ #
    # |                                    VOXELISATION                                       | #
    # +---------------------------------------------------------------------------------------+ #

    def image_from_pointcloud(
        self, pointcloud: np.ndarray, fov: Optional[Tuple[float]] = None
    ) -> np.ndarray:
        cfg = self.config.voxelisation
        particle_fov = F.get_FOV(pointcloud)
        if fov is None:
            fov = particle_fov
        else:
            # check if particle_fov inside fov
            x_min, x_max, y_min, y_max, z_min, z_max = fov
            x_min_p, x_max_p, y_min_p, y_max_p, z_min_p, z_max_p = particle_fov
            inside = (
                (x_min < x_min_p)
                and (y_min < y_min_p)
                and (z_min < z_min_p)
                and (x_max > x_max_p)
                and (y_max > y_max_p)
                and (z_max > z_max_p)
            )
            if not inside:
                raise ValueError("Particle not inside fov", fov, particle_fov)
        x_min, x_max, y_min, y_max, z_min, z_max = fov
        x_min, x_max, y_min, y_max, z_min, z_max = map(
            float, (x_min, x_max, y_min, y_max, z_min, z_max)
        )
        step = float(self.step)
        x_range = int(np.ceil((x_max - x_min) / step))
        y_range = int(np.ceil((y_max - y_min) / step))
        z_range = int(np.ceil((z_max - z_min) / step))
        target_points = np.mgrid[
            x_min : x_max : x_range * 1j,
            y_min : y_max : y_range * 1j,
            z_min : z_max : z_range * 1j,
        ]
        target_points = target_points.reshape(3, -1).T
        weights = np.ones(len(pointcloud))
        figtree_kwargs = {"bandwidth": cfg.bandwidth, "epsilon": cfg.epsilon}
        densities = figtree(
            pointcloud.astype(np.float64), target_points, weights, **figtree_kwargs
        )
        densities = densities.astype(self.dtype)
        densities = densities.reshape((x_range, y_range, z_range))
        densities = (densities - densities.min()) / (densities.max() - densities.min())

        return densities

    # +---------------------------------------------------------------------------------------+ #
    # |                                   AUGMENT IMAGE                                       | #
    # +---------------------------------------------------------------------------------------+ #

    @staticmethod
    def add_one_outliers_cluster(image: np.ndarray, cfg: OutliersConfig):
        radius_xy_min, radius_xy_max = cfg.radius_range_xy
        radius_z_min, radius_z_max = cfg.radius_range_z
        radius_x = R.randint(radius_xy_min, high=radius_xy_max)
        radius_y = R.randint(radius_xy_min, high=radius_xy_max)
        radius_z = R.randint(radius_z_min, high=radius_z_max)
        center_x = R.randint(radius_x, high=image.shape[2] - radius_x)
        center_y = R.randint(radius_y, high=image.shape[1] - radius_y)
        center_z = R.randint(radius_z, high=image.shape[0] - radius_z)
        center = (center_z, center_y, center_x)
        radius = (radius_z, radius_y, radius_x)
        low, high = cfg.nb_points_range
        nb_points = R.randint(low, high=high)
        gaussian_kwargs = {
            "loc": center,
            "scale": np.sqrt(radius),
            "size": (nb_points, 3),
        }
        outliers_cluster = R.normal(**gaussian_kwargs).astype(np.uint16)
        low, high = cfg.intensity_range
        for z, y, x in outliers_cluster:
            image[z, y, x] = (high - low) * R.random_sample() + low

    def add_outliers_clusters(self, image: np.ndarray, cfg: OutliersConfig):
        low, high = cfg.nb_clusters_range
        nb_clusters = R.randint(low, high)
        for _ in range(nb_clusters):
            try:
                self.add_one_outliers_cluster(image, cfg)
            except IndexError:
                pass

    def add_all_outliers(self):
        for one_pass_config in self.config.outliers:
            self.add_outliers_clusters(self.output_image, one_pass_config)

    def add_anisotropic_blur(self):
        sigma = self.config.sensor.anisotropic_blur_sigma
        mode = self.config.sensor.anisotropic_blur_border_mode
        self.output_image = gaussian_filter(self.output_image, sigma=sigma, mode=mode)

    def add_gaussian_noise_with_target_snr(self):
        image_average_db = 10 * np.log10(self.output_image.mean())
        noise_average_db = (
            image_average_db - self.config.sensor.gaussian_noise_target_snr_db
        )
        noise_average = 10 ** (noise_average_db / 10)
        noise = R.normal(
            self.config.sensor.gaussian_noise_mean,
            np.sqrt(noise_average),
            self.config.target_shape,
        )
        self.output_image += noise
        pixel_values_range = self.output_image.max() - self.output_image.min()
        self.output_image = (
            self.output_image - self.output_image.min()
        ) / pixel_values_range

    def add_poisson_noise(self):
        self.output_image = random_noise(self.output_image, "poisson")

    def add_sensor_faults(self):
        if self.config.sensor.anisotropic_blur:
            self.add_anisotropic_blur()
        if self.config.sensor.gaussian_noise:
            self.add_gaussian_noise_with_target_snr()
        if self.config.sensor.poisson_noise:
            self.add_poisson_noise()

    @staticmethod
    def make_psf(anisotropic_blur_sigma: Tuple[int | float], dtype: np.dtype) -> np.ndarray:
        # commented out because GT is in the image space not int the pointcloud space
        # step = float(self.step)
        # B = self.config.voxelisation.bandwidth # bandwidth in the pointcloud coordinate system
        # B = int(np.ceil(B / step)) # bandwidth in the image space
        # k = 10
        # shape = [2*k*B+1] * 3
        # psf = np.zeros(shape, dtype=float)
        # psf[k*B,k*B,k*B] = 1 # dirac
        # psf = gaussian_filter(psf, sigma=B, mode='constant') # gaussian of std B

        sigma = np.array(anisotropic_blur_sigma, dtype=float)
        k = 5  # +/- 5 sigmas is sufficient
        shape = np.ceil(2 * k * sigma + 1).astype(int)
        size = max(shape)
        grid = make_grid(size, 3)
        grid_step = 2 / (size - 1)
        cov_PSF = grid_step**2 * np.eye(3)
        cov_PSF[[0, 1, 2], [0, 1, 2]] *= sigma**2
        psf = nd_gaussian(grid, np.zeros(3), cov_PSF, 3).astype(dtype)
        if psf.max() > psf.min():
            psf = (psf - psf.min()) / (psf.max() - psf.min())
        psf = psf / psf.sum()
        return psf

    def create_psf(self):
        # Anisotropic blur
        if self.config.sensor.anisotropic_blur:
            self.psf = DataGenerator.make_psf(
                self.config.sensor.anisotropic_blur_sigma, self.dtype
            )
        else:
            self.psf = None

    # +---------------------------------------------------------------------------------------+ #
    # |                                      WRAPPERS                                         | #
    # +---------------------------------------------------------------------------------------+ #

    def save_psf(self, path: str):
        tifffile.imwrite(path, self.psf)

    def save_groundtruth(self, path: str):
        # center the pointcloud
        gt = self.image_from_pointcloud(
            self.template_pointcloud.cpu().numpy(), fov=self.common_fov
        )
        tifffile.imwrite(path, gt)

    def create_particles(self, output: str = None, output_extension: str = "npz"):
        angles = self.generate_angles(
            self.config.voxelisation.tilt_strategy, self.config.voxelisation.tilt_margin
        )
        translations = self.generate_translations(
            self.config.augmentation.max_translation
        )
        particles, real_translations = [], []

        if output is not None:
            output_dir = Path(output)
            os.makedirs(output_dir, exist_ok=True)
            particles_dir = output_dir / "particles"
            os.makedirs(particles_dir, exist_ok=True)
            f = open(output_dir / "poses.csv", "w")
            csvwriter = csv.writer(f)
            csvwriter.writerow(["name", "rot1", "rot2", "rot3", "t1", "t2", "t3"])

        for i, rotation_angles, translation in tqdm(
            zip(range(len(angles)), angles, translations),
            total=len(angles),
            desc="Creating particles...",
        ):
            particle = self.image_from_pointcloud(
                self.augment_pointcloud(
                    rotation_angles=rotation_angles, translation=translation
                ),
                fov=self.common_fov,
            )
            translation /= self.step
            translation = translation

            # add anisotropic blur
            sigma = self.config.sensor.anisotropic_blur_sigma
            mode = self.config.sensor.anisotropic_blur_border_mode
            if any([s > 0 for s in sigma]):
                ndii.convolve(particle, self.psf, mode=mode, cval=0.0, output=particle)

            # add gaussian noise
            image_average_db = 10 * np.log10(particle.mean())
            noise_average_db = (
                image_average_db - self.config.sensor.gaussian_noise_target_snr_db
            )
            noise_average = 10 ** (noise_average_db / 10)
            noise = R.normal(
                self.config.sensor.gaussian_noise_mean,
                np.sqrt(noise_average),
                particle.shape,
            )
            particle += noise
            pixel_values_range = particle.max() - particle.min()
            particle = (particle - particle.min()) / pixel_values_range

            particles.append(particle)
            real_translations.append(translation)
            if output is not None:
                csvwriter.writerow(
                    [f"particles/{i}.tiff", *rotation_angles, *translation]
                )
                tifffile.imwrite(particles_dir / f"{i}.tiff", particle)

        real_translations = np.stack(real_translations)
        if output is not None:
            f.close()

        return angles, real_translations

    def create_image(
        self, output: str = None, output_extension: str = "npz"
    ) -> np.ndarray:
        # 1. Generate centers and rotation angles for each particle, and get views from angles.
        print("Generating centers...")
        centers = self.generate_centers()
        real_centers = []
        angles = self.generate_angles(
            self.config.voxelisation.tilt_strategy, self.config.voxelisation.tilt_margin
        )
        translations = self.generate_translations(
            self.config.augmentation.max_translation
        )
        views = np.array(list(map(self.get_view_from_angles, angles)))
        # 2. Augment, voxelize and put images inside one big image
        for center, rotation_angles, translation in tqdm(
            zip(centers, angles, translations),
            total=len(centers),
            desc="Creating particles...",
        ):
            particle = self.image_from_pointcloud(
                self.augment_pointcloud(
                    rotation_angles=rotation_angles, translation=translation
                )
            )
            translation /= self.step
            depth, height, width = particle.shape
            z, y, x = center
            z_min, y_min, x_min = z - depth // 2, y - height // 2, x - width // 2
            z_max, y_max, x_max = z_min + depth, y_min + height, x_min + width
            real_center = np.array([z_min, y_min, x_min], dtype=float) + translation
            real_centers.append(real_center)
            self.output_image[z_min:z_max, y_min:y_max, x_min:x_max] += particle
        real_centers = np.stack(real_centers, axis=0)
        # 3. Add outliers and sensor faults (blur + noise(s))
        self.add_all_outliers()
        self.add_sensor_faults()
        # 4. Save
        if output is not None:
            output_path = (
                os.path.join(self.config.io.output_dir, output) + "." + output_extension
            )
            if output_extension == "npz":
                np.savez(output_path, image=self.output_image)
            elif output_extension == "tiff":
                self.output_image = self.output_image.astype(float)
                self.output_image = (self.output_image - self.output_image.min()) / (
                    self.output_image.max() - self.output_image.min()
                )
                tifffile.imwrite(
                    output_path, (255 * self.output_image).astype(np.uint8)
                )
        return real_centers, views, angles, translations

    def generate_dataset(self):
        os.makedirs(self.config.io.output_dir, exist_ok=True)
        self.create_psf()
        all_centers, all_views, all_angles = [], [], []
        for i in tqdm(range(self.config.io.generated_dataset_size)):
            centers, views, angles, _ = self.create_image(
                output=str(i), output_extension=self.config.io.extension
            )
            width = int(np.ceil(np.log10(len(centers))))
            for j, (center, view, angles_) in enumerate(zip(centers, views, angles)):
                all_centers.append(
                    (f"{i}.{self.config.io.extension}", str(j).zfill(width), *center)
                )
                all_views.append(
                    (f"{i}.{self.config.io.extension}", str(j).zfill(width), view)
                )
                all_angles.append(
                    (f"{i}.{self.config.io.extension}", str(j).zfill(width), *angles_)
                )
            self.output_image = np.zeros(self.config.target_shape)  # reset output image
        with open(f"{self.config.io.output_dir}/coordinates.csv", "w", newline="") as f:
            csv.writer(f).writerows(all_centers)
        with open(f"{self.config.io.output_dir}/views.csv", "w", newline="") as f:
            csv.writer(f).writerows(all_views)
        with open(f"{self.config.io.output_dir}/angles.csv", "w", newline="") as f:
            csv.writer(f).writerows(all_angles)
        with open(f"{self.config.io.output_dir}/config.pickle", "wb") as f:
            pickle.dump(self.config, f)



import argparse
import csv

import numpy as np
import scipy.ndimage as ndi

from .get_poses import (
    iterative_suppression,
    load_patches,
    load_views,
    normalize_patches,
)


def parse_args():
    parser = argparse.ArgumentParser("Setup pipeline")

    parser.add_argument("--crop_dir", type=str, default=None)
    parser.add_argument("--views", type=str, default=None)

    parser.add_argument("--number_side", type=int, default=None)
    parser.add_argument("--number_top", type=int, default=None)

    parser.add_argument("--poses", type=str, default=None)
    return parser.parse_args()


def main(args):
    views, image_names = load_views(args.views)
    patches = load_patches(args.crop_dir, image_names)

    patches_2D = patches.sum(axis=1)
    patches_2D = normalize_patches(patches_2D)
    (
        mask_keep_side,
        params_side,
        avrg_particle_side,
        variance_map_side,
    ) = iterative_suppression(patches_2D[views == 1], args.number_side)
    (
        mask_keep_top,
        params_top,
        avrg_particle_top,
        variance_map_top,
    ) = iterative_suppression(patches_2D[views == 0], args.number_top)

    params = np.zeros(
        (patches.shape[0], 6)
    )  # shape (N, (euler1, euler2, euler3, tz, ty, tx))
    mask_keep = np.zeros((patches.shape[0]), dtype=bool)

    # Side params + trouver l'angle absolu
    params[views == 1, [0, 4, 5]] = params_side
    params[views == 1, 1] = 90
    mask_keep[views == 1] = mask_keep_side

    # Top params
    params[views == 0, [2, 4, 5]] = params_top
    mask_keep[views == 0] = mask_keep_top

    # Z-axis translation
    # Mass center in the center of the patch
    centers_of_mass = np.array(list(map(lambda x: ndi.center_of_mass(x)[0], patches)))
    params[:, 3] = -centers_of_mass

    with open(args.poses, "w") as f:
        csv.writer(f).writerows(params)


if __name__ == "__main__":
    main(parse_args())



"""A minimalistic Pytorch implementation of EfficientNet directly based on the original paper:
"EfficientNet: Rethinking Model Scaling for Convolutional Neural Networks",
Mingxing Tan, Quoc V. Le, 2019, https://arxiv.org/abs/1905.11946
It allows 2D or 3D inputs thanks to the small LayerNd functions.
A small Decoder is also provided in order to use EfficientNet as the backbone for an
autoencoder.
"""

import math

import torch
from torch import nn

# +------------------------------------------------------------------------------------------+ #
# |                                         ND LAYERS                                        | #
# +------------------------------------------------------------------------------------------+ #


def ConvNd(*args, dim=2, **kwargs):
    return nn.Conv2d(*args, **kwargs) if dim == 2 else nn.Conv3d(*args, **kwargs)


def BatchNormNd(*args, dim=2, **kwargs):
    return (
        nn.BatchNorm2d(*args, **kwargs) if dim == 2 else nn.BatchNorm3d(*args, **kwargs)
    )


def AdaptativeAvgPoolNd(*args, dim=2, **kwargs):
    if dim == 2:
        return nn.AdaptiveAvgPool2d(*args, **kwargs)
    return nn.AdaptiveAvgPool3d(*args, **kwargs)


def ConvTransposeNd(*args, dim=2, **kwargs):
    if dim == 2:
        return nn.ConvTranspose2d(*args, **kwargs)
    return nn.ConvTranspose3d(*args, **kwargs)


# +------------------------------------------------------------------------------------------+ #
# |                                      BUILDING BLOCKS                                     | #
# +------------------------------------------------------------------------------------------+ #


class ConvBnAct(nn.Sequential):
    """Layer grouping a convolution, batchnorm, and activation function.

    Quoting original paper Section 5.2: "We train our EfficientNet models with [...] batch norm
    momentum 0.99 [...]. We also use SiLU (Swish-1) activation (Ramachandran et al., 2018;
    Elfwing et al., 2018; Hendrycks & Gimpel, 2016).
    """

    def __init__(
        self,
        n_in,
        n_out,
        kernel_size=3,
        stride=1,
        padding=0,
        groups=1,
        bias=False,
        bn=True,
        act=True,
        dim=2,
    ):
        super().__init__()
        conv_params = {
            "kernel_size": kernel_size,
            "stride": stride,
            "padding": padding,
            "groups": groups,
            "bias": bias,
        }
        self.add_module("conv", ConvNd(n_in, n_out, dim=dim, **conv_params))
        self.add_module(
            "bn", BatchNormNd(n_out, dim=dim, momentum=0.99) if bn else nn.Identity()
        )
        self.add_module("act", nn.SiLU() if act else nn.Identity())


class SEBlock(nn.Module):
    """Squeeze-and-excitation block."""

    def __init__(self, n_in, r=24, dim=2):
        super().__init__()
        self.squeeze = AdaptativeAvgPoolNd(1, dim=dim)
        self.excitation = nn.Sequential(
            ConvNd(n_in, n_in // r, dim=dim, kernel_size=1),
            nn.SiLU(),
            ConvNd(n_in // r, n_in, dim=dim, kernel_size=1),
            nn.Sigmoid(),
        )

    def forward(self, x):
        return x * self.excitation(self.squeeze(x))


class DropSample(nn.Module):
    """Drops each sample in x with probability p during training (a sort of DropConnect).

    In the original paper Dropout regularization is mentionned but it's not what the official
    repo shows. See this discussion: https://github.com/tensorflow/tpu/issues/494.

    Official tensorflow code:
    https://github.com/tensorflow/tpu/blob/master/models/official/efficientnet/utils.py#L276
    """

    def __init__(self, p=0, dim=2):
        super().__init__()
        self.p = p
        self.size = (dim + 1) * (1,)

    def forward(self, x):
        if (not self.p) or (not self.training):
            return x
        random_tensor = torch.FloatTensor(len(x), *self.size).uniform_().to(x.device)
        bit_mask = self.p < random_tensor
        x = x.div(1 - self.p)
        x = x * bit_mask
        return x


# +------------------------------------------------------------------------------------------+ #
# |                                        MBCONV BLOCK                                      | #
# +------------------------------------------------------------------------------------------+ #


class MBConvN(nn.Module):
    """MBConv with an expansion factor of N, plus squeeze-and-excitation.

    Replace costly 3 X 3 convolutions with depthwise convolutions and follows a
    narrow -> wide -> narrow structure as opposed to the wide -> narrow -> wide one found in
    original residual blocks.

    Quoting original paper Section 4: "Its main building block is mobile inverted bottleneck
    MBConv (Sandler et al., 2018; Tan et al., 2019), to which we also add squeeze-and-excitation
    optimization (Hu et al., 2018)."
    """

    def __init__(
        self, n_in, n_out, expand_factor, kernel_size=3, stride=1, r=24, p=0, dim=2
    ):
        super().__init__()

        expanded = expand_factor * n_in
        padding = (kernel_size - 1) // 2
        depthwise_conv_params = {
            "kernel_size": kernel_size,
            "padding": padding,
            "stride": stride,
            "groups": expanded,
        }

        self.skip_connection = (n_in == n_out) and (stride == 1)

        if expand_factor == 1:
            self.expand_pw = nn.Identity()
        else:
            self.expand_pw = ConvBnAct(n_in, expanded, kernel_size=1, dim=dim)
        self.depthwise = ConvBnAct(expanded, expanded, **depthwise_conv_params, dim=dim)
        self.se = SEBlock(expanded, r=r, dim=dim)
        self.reduce_pw = ConvBnAct(expanded, n_out, kernel_size=1, act=False, dim=dim)
        self.dropsample = DropSample(p, dim)

    def forward(self, x):
        residual = x
        x = self.reduce_pw(self.se(self.depthwise(self.expand_pw(x))))
        if self.skip_connection:
            x = self.dropsample(x)
            x = x + residual
        return x


class MBConv1(MBConvN):
    def __init__(self, n_in, n_out, kernel_size=3, stride=1, r=24, p=0, dim=2):
        super().__init__(
            n_in,
            n_out,
            expand_factor=1,
            kernel_size=kernel_size,
            stride=stride,
            r=r,
            p=p,
            dim=dim,
        )


class MBConv6(MBConvN):
    def __init__(self, n_in, n_out, kernel_size=3, stride=1, r=24, p=0, dim=2):
        super().__init__(
            n_in,
            n_out,
            expand_factor=6,
            kernel_size=kernel_size,
            stride=stride,
            r=r,
            p=p,
            dim=dim,
        )


# +------------------------------------------------------------------------------------------+ #
# |                                       EFFICIENT NET                                      | #
# +------------------------------------------------------------------------------------------+ #


class EfficientNet(nn.Module):
    """Generic EfficientNet that takes in the width and depth scale factors and scales
    accordingly.
    In addition to the standard b0 to b7 configuration, it allows two new config, namely
    's' and 'xs'. Those are almost equals to b0 except for one or two stride(s), respectively
    for 's' and 'xs', that are 1 instead of 2 in order to allow smaller inputs.
    It can handle 2D or 3D data through the 'dim' argument.
    """

    def __init__(
        self,
        in_channels=1,
        width_factor=1,
        depth_factor=1,
        num_classes=1000,
        mode=None,
        dim=2,
    ):
        super().__init__()

        base_widths = [32, 16, 24, 40, 80, 112, 192, 320, 1280]
        base_depths = [1, 2, 2, 3, 3, 4, 1]
        kernel_sizes = [3, 3, 5, 3, 5, 5, 3]
        strides = [1, 2, 2, 2, 1, 2, 1]
        if mode == "s" or mode == "xs":
            strides[1] = 1
        drop_probas = [0, 0.029, 0.057, 0.086, 0.114, 0.143, 0.171]
        reduce = [4] + 6 * [24]
        layers = [MBConv1] + 6 * [MBConv6]

        scaled_widths = [self.scale_width(w, width_factor) for w in base_widths]
        scaled_depths = [math.ceil(depth_factor * d) for d in base_depths]

        stage_params = [
            (
                scaled_widths[i],
                scaled_widths[i + 1],
                scaled_depths[i],
                layers[i],
                kernel_sizes[i],
                strides[i],
                reduce[i],
                drop_probas[i],
                dim,
            )
            for i in range(7)
        ]

        # Stage 1 in the original paper Table 1
        stem_params = {"stride": 1 if mode == "xs" else 2, "padding": 1, "dim": dim}
        self.stem = ConvBnAct(in_channels, scaled_widths[0], **stem_params)

        # Stages 2 to 7 in the original paper Table 1
        self.stages = nn.Sequential(
            *[self.create_stage(*stage_params[i]) for i in range(7)]
        )

        # First layer of stage 9 in the original paper Table 1
        self.pre_head = ConvBnAct(
            scaled_widths[-2], scaled_widths[-1], kernel_size=1, dim=dim
        )

        # Second and third layers of stage 9 in the original paper Table 1
        self.head = nn.Sequential(
            AdaptativeAvgPoolNd(1, dim=dim),
            nn.Flatten(),
            nn.Linear(scaled_widths[-1], num_classes),
        )

        self.latent_dim = scaled_widths[-1]

    @staticmethod
    def create_stage(
        n_in, n_out, num_layers, layer_type, kernel_size, stride, r, p, dim
    ):
        """Creates a Sequential consisting of [num_layers] layer_type."""
        common_params = {"kernel_size": kernel_size, "r": r, "p": p, "dim": dim}
        layers = [layer_type(n_in, n_out, stride=stride, **common_params)]
        layers += [
            layer_type(n_out, n_out, **common_params) for _ in range(num_layers - 1)
        ]
        return nn.Sequential(*layers)

    @staticmethod
    def scale_width(w, w_factor):
        """Scales width given a scale factor.
        See:
        https://stackoverflow.com/questions/60583868/how-is-the-number-of-channels-adjusted-in-efficientnet.
        """
        w *= w_factor
        new_w = (int(w + 4) // 8) * 8
        new_w = max(8, new_w)
        if new_w < 0.9 * w:
            new_w += 8
        return int(new_w)

    def extract_features(self, x):
        return self.pre_head(self.stages(self.stem(x)))

    def forward(self, x):
        return self.head(self.extract_features(x))

    @property
    def num_params(self):
        return sum(p.numel() for p in self.parameters() if p.requires_grad)

    @classmethod
    def from_name(cls, network_name, **kwargs):
        """Official tensorflow code:
        https://github.com/tensorflow/tpu/blob/master/models/official/efficientnet/efficientnet_builder.py#L35
        """
        config = {
            "efficientnet-b0": {"width_factor": 1.0, "depth_factor": 1.0},
            "efficientnet-b1": {"width_factor": 1.0, "depth_factor": 1.1},
            "efficientnet-b2": {"width_factor": 1.1, "depth_factor": 1.2},
            "efficientnet-b3": {"width_factor": 1.2, "depth_factor": 1.4},
            "efficientnet-b4": {"width_factor": 1.4, "depth_factor": 1.8},
            "efficientnet-b5": {"width_factor": 1.6, "depth_factor": 2.2},
            "efficientnet-b6": {"width_factor": 1.8, "depth_factor": 2.6},
            "efficientnet-b7": {"width_factor": 2.0, "depth_factor": 3.1},
            # b0 with smaller inputs
            "efficientnet-s": {"width_factor": 1.0, "depth_factor": 1.0, "mode": "s"},
            "efficientnet-xs": {"width_factor": 1.0, "depth_factor": 1.0, "mode": "xs"},
        }
        return cls(**{**config[network_name], **kwargs})


# +------------------------------------------------------------------------------------------+ #
# |                                          DECODER                                         | #
# +------------------------------------------------------------------------------------------+ #


class Decoder(nn.Module):
    """A simple Decoder Module that can be used with EfficientNet as an encoder, e.g:
    > dim  = 2  # can be 2 or 3
    > size = 24
    > encoder = EfficientNet.from_name('efficientnet-s', dim=dim)
    > decoder = Decoder(out_size=size, dim=dim, latent_dim=encoder.latent_dim)
    > x = torch.randn(1, 1, *(dim * (size, )))
    > x_hat = decoder(encoder.extract_features(x))
    """

    def __init__(
        self,
        out_channels=1,
        out_size=24,
        dim=2,
        hidden_dim=128,
        num_features=(8, 16, 32),
        latent_dim=4,
    ) -> None:
        super().__init__()

        self.out_size = out_size
        unflattened_size = (num_features[-1], *(dim * (3,)))

        first_conv_params = {"kernel_size": 3, "stride": 2, "output_padding": 0}
        other_conv_params = {
            "kernel_size": 3,
            "stride": 2,
            "output_padding": 1,
            "padding": 1,
        }

        self.linear = nn.Sequential(
            AdaptativeAvgPoolNd(1, dim=dim),
            nn.Flatten(),
            nn.Linear(latent_dim, hidden_dim),
            nn.ReLU(inplace=True),
            nn.Linear(hidden_dim, torch.as_tensor(unflattened_size).prod().item()),
            nn.ReLU(inplace=True),
        )

        self.unflatten = nn.Unflatten(dim=1, unflattened_size=unflattened_size)

        self.conv = nn.Sequential(
            ConvTransposeNd(
                num_features[-1], num_features[-2], dim=dim, **first_conv_params
            ),
            BatchNormNd(num_features[-2], dim=dim),
            nn.ReLU(True),
            ConvTransposeNd(
                num_features[-2], num_features[-3], dim=dim, **other_conv_params
            ),
            BatchNormNd(num_features[-3], dim=dim),
            nn.ReLU(True),
            ConvTransposeNd(
                num_features[-3], out_channels, dim=dim, **other_conv_params
            ),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # linear -> unflatten -> conv -> resize -> sigmoid
        x = self.conv(self.unflatten(self.linear(x)))
        x = nn.functional.interpolate(x, size=self.out_size)
        return x.sigmoid()

    @property
    def num_params(self):
        return sum(p.numel() for p in self.parameters() if p.requires_grad)



from __future__ import annotations

import os
from typing import TYPE_CHECKING

import numpy as np

from spfluo.utils.array import array_namespace, to_numpy
from spfluo.utils.volume import center_of_mass, fourier_shift

from ...utils.read_save_files import save_image
from ..common_image_processing_methods.others import crop_center
from ..common_image_processing_methods.registration import (
    center_connected_component,
    registration_exhaustive_search,
    shift_registration_exhaustive_search,
)

if TYPE_CHECKING:
    from spfluo.utils.array import Array


class Fourier_pixel_representation:
    def __init__(
        self,
        nb_dim: int,
        size: tuple[int, ...],
        psf: "Array",
        init_vol: "Array" | None = None,
        random_init: bool = True,
        dtype: str = "float32",
    ):
        xp = array_namespace(psf)
        self.xp = xp
        self.dtype = getattr(xp, dtype)
        complex_type_promotion = {
            "float32": "complex64",
            "float64": "complex128",
        }
        self.complex_dtype = getattr(xp, complex_type_promotion[dtype])
        if init_vol is None:
            if random_init:
                volume_fourier = np.random.randn(*psf.shape) + 1j * np.random.randn(
                    *psf.shape
                )
            else:
                raise NotImplementedError
        else:
            volume_fourier = np.fft.fftn(
                np.fft.ifftshift(crop_center(np.asarray(init_vol), (size, size, size)))
            )
        self.volume_fourier = xp.asarray(volume_fourier, dtype=self.complex_dtype)
        self.nb_dim = nb_dim
        self.size = size
        self.psf = xp.asarray(psf, dtype=self.dtype)
        self.psf /= xp.sum(self.psf)

    def gd_step(self, grad, lr, reg_coeff=0):
        self.volume_fourier -= lr * grad
        gradient_l2_reg, l2_reg = self.l2_regularization()
        self.volume_fourier -= lr * reg_coeff * gradient_l2_reg

    def l2_regularization(self):
        gradient_l2_reg = 2 * self.volume_fourier
        l2_reg = self.xp.mean(self.xp.abs(self.volume_fourier) ** 2)
        return gradient_l2_reg, l2_reg

    def get_image_from_fourier_representation(self):
        ifft = self.xp.fft.ifftn(self.volume_fourier)
        image = self.xp.abs(self.xp.fft.fftshift(ifft)).real
        return image

    def save(self, output_dir, output_name):
        path = f"{output_dir}/{output_name}.ome.tiff"
        save_image(path, self.get_image_from_fourier_representation(), order="ZYX")

    def register_and_save(self, output_dir, output_name, ground_truth=None):
        im = self.get_image_from_fourier_representation()

        if ground_truth is not None:
            _, im = shift_registration_exhaustive_search(ground_truth, im)
            im = im.astype(ground_truth.dtype)
            _, im = registration_exhaustive_search(ground_truth, im)

        path = os.path.join(output_dir, output_name)
        save_image(path, im, order="ZYX")
        return im

    def center(self):
        vol, shift = center_connected_component(
            to_numpy(self.get_image_from_fourier_representation())
        )
        shift = self.xp.asarray(shift)
        com = self.xp.asarray(center_of_mass(self.xp.asarray(vol)))
        c = (self.xp.asarray(self.volume_fourier.shape) - 1) / 2
        shift += c - com
        self.volume_fourier = self.xp.astype(
            fourier_shift(self.volume_fourier, shift), self.complex_dtype
        )
        return shift



from time import time

import numpy as np
from matplotlib.widgets import Slider
from scipy.interpolate import UnivariateSpline


def get_rotation_mat(angle):
    angle = float(angle)
    return np.array(
        [
            [np.cos(angle), -np.sin(angle), 0],
            [np.sin(angle), np.cos(angle), 0],
            [0, 0, 1],
        ]
    )


def simu_cylinder_surface_random(a, b, L, N):
    angle = np.linspace(0, 2 * np.pi, num=N, endpoint=False)
    x = a * np.cos(angle)
    y = b * np.sin(angle)
    z = L * np.random.random(size=(N,))
    # z = L * np.linspace(0, 1, num=N)
    return np.stack([x, y, z], axis=1)


def simu_triplet_random(r, L, s, nbPts):
    n12 = nbPts // 3
    n3 = nbPts - 2 * n12
    V1 = simu_cylinder_surface_random(r, r, L, n12)
    V2 = simu_cylinder_surface_random(r, r, L, n12)
    V3 = simu_cylinder_surface_random(r, r, L, n3)
    V2[:, 0] += s
    V3[:, 0] += 2 * s
    return np.concatenate([V1, V2, V3], axis=0)


def simu_cylinder_rev(radius, r, L, s, nbPts, angle, triplet=1):
    Cn = 9

    angle = angle - 180
    # t0 = time()
    # Radius variation profile
    x0 = np.linspace(0, L, num=len(radius))
    y0 = radius
    radius_profile_spl = UnivariateSpline(x0, y0, s=1)
    # t1 = time()

    samplingThetaCn = 2 * np.pi / Cn
    if triplet == 1:
        Vi0 = simu_triplet_random(r, L, s, nbPts)
    else:
        Vi0 = simu_cylinder_surface_random(r * 1, r * 0.2, L, nbPts)

    # t2 = time()

    angle = angle * np.pi / 180
    R = get_rotation_mat(angle)
    Vi0 = (R @ Vi0[:, :, None])[:, :, 0]

    V = np.empty((nbPts, Cn, 3))
    # t3 = time()
    for i in range(Cn):
        angle = i * samplingThetaCn
        R = get_rotation_mat(angle)
        V[:, i] = (R @ Vi0[:, :, None])[:, :, 0]

        Dinner = radius_profile_spl(V[:, i, 2])
        V[:, i, 0] += (Dinner + r) * np.cos(angle)
        V[:, i, 1] += (Dinner + r) * np.sin(angle)
    # t4 = time()

    # print(f"Spline calc : {t1-t0:.6f}s")
    # print(f"Vi0 calc : {t2-t1:.6f}s")
    # print(f"Rotation : {t3-t2:.6f}s")
    # print(f"V calc : {t4-t3:.6f}s")

    return V.reshape(-1, 3)


def get_radius(x, a, b, c):
    # get the radius
    return np.round(a * (x - b) ** 2 + c)


if __name__ == "__main__":
    np.random.seed(0)

    x = np.arange(1, 14)
    a = -0.5  # realistic range: [-0.535,0]
    b = 5.4  # realistic range: [5.4,6]
    c = 103.253  # realistic range: [103.253,105]

    r = 15  # r diameter of the microtubules of the triplets
    L = 600  # l Length of the barrel
    s = 15  # s Length between microtubules of the triplet
    nbPts = 1000
    angle = 110  # angle of the triplets

    V = simu_cylinder_rev(get_radius(x, a, b, c), r, L, s, nbPts, angle)

    import matplotlib.pyplot as plt

    fig = plt.figure()
    ax = fig.add_subplot(projection="3d")

    # adjust the main plot to make room for the sliders
    plt.subplots_adjust(bottom=0.25)

    # Make a horizontal slider to control a, b, c.
    a_slider = Slider(
        ax=plt.axes([0.1, 0.2, 0.8, 0.01]),
        label="a",
        valmin=-0.535,
        valmax=0,
        valinit=a,
    )

    b_slider = Slider(
        ax=plt.axes([0.1, 0.18, 0.8, 0.01]),
        label="b",
        valmin=5.4,
        valmax=6,
        valinit=b,
    )

    c_slider = Slider(
        ax=plt.axes([0.1, 0.16, 0.8, 0.01]),
        label="c",
        valmin=103.253,
        valmax=105,
        valinit=c,
    )

    r_slider = Slider(
        ax=plt.axes([0.1, 0.14, 0.8, 0.01]),
        label="r",
        valmin=10,
        valmax=20,
        valinit=r,
    )

    l_slider = Slider(
        ax=plt.axes([0.1, 0.12, 0.8, 0.01]),
        label="l",
        valmin=10,
        valmax=1000,
        valinit=L,
    )

    s_slider = Slider(
        ax=plt.axes([0.1, 0.10, 0.8, 0.01]),
        label="s",
        valmin=10,
        valmax=20,
        valinit=s,
    )

    angle_slider = Slider(
        ax=plt.axes([0.1, 0.08, 0.8, 0.01]),
        label="angle",
        valmin=0,
        valmax=180,
        valinit=angle,
    )

    X, Y, Z = V[:, 0], V[:, 1], V[:, 2]
    sc = ax.scatter(X, Y, Z, s=0.1)

    t0 = time()

    def update(val):
        global t0
        radius = get_radius(x, a_slider.val, b_slider.val, c_slider.val)
        V = simu_cylinder_rev(
            radius, r_slider.val, l_slider.val, s_slider.val, nbPts, angle_slider.val
        )
        sc._offsets3d = (V[:, 0], V[:, 1], V[:, 2])
        t = time()
        print(f"FPS : {1/(t-t0):.2f}")
        t0 = t
        # fig.canvas.draw_idle()

    a_slider.on_changed(update)
    b_slider.on_changed(update)
    c_slider.on_changed(update)
    r_slider.on_changed(update)
    l_slider.on_changed(update)
    s_slider.on_changed(update)
    angle_slider.on_changed(update)

    max_range = (
        np.array([X.max() - X.min(), Y.max() - Y.min(), Z.max() - Z.min()]).max() / 2.0
    )

    mid_x = (X.max() + X.min()) * 0.5
    mid_y = (Y.max() + Y.min()) * 0.5
    mid_z = (Z.max() + Z.min()) * 0.5
    ax.set_xlim(mid_x - max_range, mid_x + max_range)
    ax.set_ylim(mid_y - max_range, mid_y + max_range)
    ax.set_zlim(mid_z - max_range, mid_z + max_range)

    plt.show()
